import os, sys
import numpy as np
import sys
from math import *
import bisect
from astropy.time import Time
import astropy.units as u
import warnings
from suncasa.io import ndfits
import ssl
from scipy.interpolate import interp1d

py3 = sys.version_info.major >= 3
if py3:
    ## CASA version >= 6
    from urllib.request import urlopen
else:
    ## CASA version < 6
    from urllib2 import urlopen

from ..casa_compat import import_casatools
tools = import_casatools(['tbtool', 'mstool', 'qatool', 'iatool'])

tbtool = tools['tbtool']
mstool = tools['mstool']
qatool = tools['qatool']
iatool = tools['iatool']
tb = tbtool()
ms = mstool()
qa = qatool()
ia = iatool()


import sunpy

# check sunpy version
sunpy1 = sunpy.version.major >= 1
if sunpy1:
    from sunpy.coordinates import sun
else:
    from sunpy import sun

try:
    from astropy.io import fits as pyfits
except:
    try:
        import pyfits
    except ImportError:
        raise ImportError('Neither astropy nor pyfits exists in this CASA installation')


def ms_clearhistory(msfile):
    tb_history = msfile + '/HISTORY'
    os.system('cp -r {0} {0}_bk'.format(tb_history))
    tb.open(tb_history, nomodify=False)
    nrows = tb.nrows()
    if nrows > 0:
        tb.removerows(range(nrows))
    tb.close()


def normalize(angle, lower=-np.pi, upper=np.pi):
    """
    Function to normalize any angles to between the lower limit and upper limit
    :param angle: input angle in radians
    :param lower: lower bound of the normalization, default to -pi
    :param upper: upper bound of the normalization, default to pi
    :return: normalized angle in radians
    """
    if lower >= upper:
        raise ValueError("Invalid lower and upper limits: (%s, %s)" %
                         (lower, upper))

    res = angle
    if angle > upper or angle == lower:
        angle = lower + abs(angle + upper) % (abs(lower) + abs(upper))
    if angle < lower or angle == upper:
        angle = upper - abs(angle - lower) % (abs(lower) + abs(upper))

    res = lower if res == upper else angle
    return res


def ms_restorehistory(msfile):
    tb_history = msfile + '/HISTORY'
    os.system('mv {0}_bk {0}'.format(tb_history))


def read_horizons(t0=None, dur=None, vis=None, observatory=None, verbose=False):
    """
    This function visits JPL Horizons to retrieve J2000 topocentric RA and DEC of the solar disk center
    as a function of time.

    Keyword arguments:
    t0: Referece time in astropy.Time format
    dur: duration of the returned coordinates in days. Default to 1 minute
    vis: CASA visibility dataset (in measurement set format). If provided, use entire duration from
         the visibility data
    observatory: observatory code (from JPL Horizons). If not provided, use information from visibility.
         if no visibility found, use earth center (code=500)
    verbose: True to provide extra information

    Usage:
    >>> from astropy.time import Time
    >>> out = read_horizons(t0=Time('2017-09-10 16:00:00'), observatory='-81')
    >>> out = read_horizons(vis = 'mydata.ms')

    History:
    BC (sometime in 2014): function was first wrote, followed by a number of edits by BC and SY
    BC (2019-07-16): Added docstring documentation

    '''
    
=======
    """
    if not t0 and not vis:
        t0 = Time.now()
    if not dur:
        dur = 1. / 60. / 24.  # default to 1 minute
    if t0:
        try:
            btime = Time(t0)
            
        except:
            print('input time ' + str(t0) + ' not recognized')
            return -1

    if observatory:
        # If "observatory" is provided, it takes precedence
        # turn observatory names into JPL Horizons' codes
        if observatory == 'EVLA' or observatory == '-5' or observatory=='VLA':
            observatory = '-5'
        elif observatory == 'EOVSA' or observatory == 'FASR' or observatory == 'OVRO_MMA' or observatory == '-81':
            observatory = '-81'
        elif observatory == 'ALMA' or observatory == '-7':
            observatory = '-7'
        elif observatory == 'GMRT' or observatory == 'uGMRT' or observatory == '399':
            observatory = '399'
        elif observatory == 'geocentric' or observatory == '500':
            observatory = '500'
        else:
            print('Observatory {} not recognized. Assume geocentric.'.format(observatory))
            observatory = '500'

    if vis:
        if not os.path.exists(vis):
            print('Input ms data ' + vis + ' does not exist! ')
            return -1
        try:
            if not observatory:
                ms.open(vis)
                metadata = ms.metadata()
                if metadata.observatorynames()[0] == 'EVLA':
                    observatory = '-5'
                elif metadata.observatorynames()[0] == 'EOVSA' or metadata.observatorynames()[0] == 'FASR' or metadata.observatorynames()[0] == 'OVRO_MMA':
                    observatory = '-81'
                elif metadata.observatorynames()[0] == 'GMRT' or metadata.observatorynames()[0] == 'uGMRT':
                    observatory = '399'
                elif metadata.observatorynames()[0] == 'ALMA':
                    observatory = '-7'
                else:
                    print('Observatory {} not recognized. Assume geocentric.'.format(metadata.observatorynames()[0]))
                    observatory = '500'
                metadata.close()
                ms.close()

            tb.open(vis)
            btime_vis = Time(tb.getcell('TIME', 0) / 24. / 3600., format='mjd')
            etime_vis = Time(tb.getcell('TIME', tb.nrows() - 1) / 24. / 3600., format='mjd')
            tb.close()
            if verbose:
                print("Beginning time of this scan " + btime_vis.iso)
                print("End time of this scan " + etime_vis.iso)

            # extend the start and end time for jpl horizons by 0.5 hr on each end
            btime = Time(btime_vis.mjd - 0.5 / 24., format='mjd')
            dur = etime_vis.mjd - btime_vis.mjd + 1.0 / 24.
        except:
            print('error in reading ms file: ' + vis + ' to obtain the ephemeris!')
            return -1

    # Neither observatory is defined or found through provided visibility. Assume geocentric.
    if not observatory:
        observatory = '500'

    # default the observatory to geocentric, if none provided
    etime = Time(btime.mjd + dur, format='mjd')
    
    try:
        cmdstr = "https://ssd.jpl.nasa.gov/api/horizons.api?format=text&TABLE_TYPE='OBSERVER'&QUANTITIES='1,17,20'&CSV_FORMAT='YES'&ANG_FORMAT='DEG'&CAL_FORMAT='BOTH'&SOLAR_ELONG='0,180'&CENTER='{}@399'&COMMAND='sun'&START_TIME='".format(
            observatory) + btime.iso.replace(' ', ',') + "'&STOP_TIME='" + etime.iso[:-4].replace(' ',
                                                                                                  ',') + "'&STEP_SIZE='1m'&SKIP_DAYLT='NO'&EXTRA_PREC='YES'&APPARENT='REFRACTED'"
        # cmdstr = "https://ssd.jpl.nasa.gov/horizons_batch.cgi?batch=1&TABLE_TYPE='OBSERVER'&QUANTITIES='1,17,20'&CSV_FORMAT='YES'&ANG_FORMAT='DEG'&CAL_FORMAT='BOTH'&SOLAR_ELONG='0,180'&CENTER='{}@399'&COMMAND='10'&START_TIME='".format(
        #     observatory) + btime.iso.replace(' ', ',') + "'&STOP_TIME='" + etime.iso[:-4].replace(' ',
        #                                                                                           ',') + "'&STEP_SIZE='1m'&SKIP_DAYLT='NO'&EXTRA_PREC='YES'&APPARENT='REFRACTED'"
        cmdstr = cmdstr.replace("'", "%27")
        # print('1################')
        
        try:
            context = ssl._create_unverified_context()
            f = urlopen(cmdstr, context=context)
        except:
            f = urlopen(cmdstr)
        lines = f.readlines()
        f.close()
    except:
        # todo use geocentric coordinate for the new VLA data
        print ("here")
        import requests, collections
        params = collections.OrderedDict()
        # params['batch'] = '1'
        # params['TABLE_TYPE'] = "'OBSERVER'"
        # params['QUANTITIES'] = "'1,17,20'"
        # params['CSV_FORMAT'] = "'YES'"
        # params['ANG_FORMAT'] = "'DEG'"
        # params['CAL_FORMAT'] = "'BOTH'"
        # params['SOLAR_ELONG'] = "'0,180'"
        # if observatory == '500':
        #     params['CENTER'] = "'500'"
        # else:
        #     params['CENTER'] = "'{}@399'".format(observatory)
        # params['COMMAND'] = "'10'"
        # params['START_TIME'] = "'{}'".format(btime.iso[:-4].replace(' ', ','))
        # params['STOP_TIME'] = "'{}'".format(etime.iso[:-4].replace(' ', ','))
        # params['STEP_SIZE'] = "'1m'"
        # params['SKIP_DAYLT'] = "'NO'"
        # params['EXTRA_PREC'] = "'YES'"
        # params['APPAENT'] = "'REFRACTED'"
        # results = requests.get("https://ssd.jpl.nasa.gov/horizons_batch.cgi", params=params)

        params['EPHEM_TYPE'] = "'OBSERVER'"
        params['QUANTITIES'] = "'1,17,20'"
        params['CSV_FORMAT'] = "'YES'"
        params['ANG_FORMAT'] = "'DEG'"
        params['CAL_FORMAT'] = "'BOTH'"
        params['SOLAR_ELONG'] = "'0,180'"
        if observatory == '500':
            params['CENTER'] = "'500'"
        else:
            params['CENTER'] = "'{}@399'".format(observatory)
        params['COMMAND'] = "'sun'"
        params['START_TIME'] = "'{}'".format(btime.iso[:-4].replace(' ', ','))
        params['STOP_TIME'] = "'{}'".format(etime.iso[:-4].replace(' ', ','))
        params['STEP_SIZE'] = "'1m'"
        params['SKIP_DAYLT'] = "'NO'"
        params['EXTRA_PREC'] = "'YES'"
        params['APPARENT'] = "'REFRACTED'"
        results = requests.get("https://ssd.jpl.nasa.gov/api/horizons.api?format=text", params=params)
        # print('2################')
        # print(results)
        lines = [ll for ll in results.iter_lines()]

    # add a check for python 3
    if py3:
        lines = [l.decode('utf-8', 'backslashreplace') for l in lines]

    #print(lines)
    nline = len(lines)
    istart = 0
    for i in range(nline):
        if lines[i][0:5] == '$$SOE':  # start recording
            istart = i + 1
        if lines[i][0:5] == '$$EOE':  # end recording
            iend = i
    newlines = lines[istart:iend]
    nrec = len(newlines)
    ephem_ = []
    t = []
    ra = []
    dec = []
    p0 = []
    delta = []
    for line in newlines:
        items = line.split(',')
        t.append(Time(float(items[1]), format='jd').mjd)
        ra.append(np.radians(float(items[4])))
        dec.append(np.radians(float(items[5])))
        p0.append(float(items[6]))
        delta.append(float(items[8]))
    # convert list of dictionary to a dictionary of arrays
    ephem = {'time': t, 'ra': ra, 'dec': dec, 'p0': p0, 'delta': delta}
    return ephem


def read_msinfo(vis=None, msinfofile=None, interp_to_scan=False, verbose=False):
    """
    Module ot read CASA measurement set and return RA, DEC, and time stamps of the phase center.
    Options for returning those from the emphemeris table (if available) or use those from the FIELD table.
    Parameters
    ----------
    vis: (required) path to the input CASA measurement set
    msinfofile: (optional) path/name of the saved numpy .npz file
    #use_ephem: (optional) if True (default), use the enclosed emphemeris table, otherwise use RA/DEC of the FIELD table
    interp_to_scan: (optional) if True, the entries of the emphemeris table
        are interpolated to the beginning of each scan. This is mainly for backward compatibility. Default is False.
    Returns
    -------
    msinfo: A dictionary contains necessary information for ephem_to_helio()
        vis: CASA measurement set
        observatory: Name of the observatory. May be used if they default to phase at the solar center.
        scans: summary info of the scans
        fieldids: a list of all FIELD ids
        scan_start_times: list of the start times of all scans, in mjd
        scan_end_times: list of the end times of all scans, in mjd
        btimes: time stamps used by ephem_to_helio() to determine the image/phase center shifts, in mjd
        ras: corresponding RA coordinates used by ephem_to_helio() to determine the image/phase center shifts, in rad
        decs: corresponding DEC coordinates used by ephem_to_helio() to determine the image/phase center shifts, in rad
    """
    import glob
    # read MS information #
    msinfo = dict.fromkeys(['vis', 'scans', 'fieldids', 'btimes', 'btimestr', 'inttimes', 'ras', 'decs', 'observatory'])
    ms.open(vis)
    metadata = ms.metadata()
    observatory = metadata.observatorynames()[0]
    if verbose:
        print('This measurement set has these fields: ', metadata.fieldnames())
    scans = ms.getscansummary()
    scanids = sorted(scans.keys(), key=lambda x: int(x))
    btimes_scan = []
    etimes_scan = []
    fieldids = []
    inttimes = []
    dirs = []
    ras_scan = []
    decs_scan = []

    for idx, scanid in enumerate(scanids):
        btimes_scan.append(scans[scanid]['0']['BeginTime'])
        etimes_scan.append(scans[scanid]['0']['EndTime'])
        fieldid = scans[scanid]['0']['FieldId']
        fieldids.append(fieldid)
        inttimes.append(scans[scanid]['0']['IntegrationTime'])
        dir = ms.getfielddirmeas('PHASE_DIR', fieldid)
        dirs.append(dir)
        ras_scan.append(dir['m0'])
        decs_scan.append(dir['m1'])

    # TODO: a caveat is that here we assume that there is only one ephemeris object in the ms
    if metadata.nfields() > 1:
        print('Warning!!! Multiple fields found in this measurement set. Will only select the first one!')

    metadata.close()
    ms.close()
    ephem_table = glob.glob(vis + '/FIELD/EPHEM*.tab')
    if len(ephem_table) == 1:
        if verbose:
            print('Found ephemeris table {}. If VLA, this is probably post-2014 data.'.format(ephem_table[0]))
            print('The visibility phase center is probably synchronized with the ephemeris. Loading data...')
        tb.open(ephem_table[0])
        col_ra = tb.getcol('RA')
        col_dec = tb.getcol('DEC')
        col_mjd = tb.getcol('MJD')
        tb.close()
        # check if the ephemeris has time that are 1 hour earlier than the measurement set
        idx, = np.where((col_mjd < (btimes_scan[0] - 1. / 24.)) | ((col_mjd > (etimes_scan[-1] + 1. / 24.))))
        if len(idx) > 0:
            print('Warning!!! I found that the ephemeris table has times that are outside that of the ms data.')
            print('Select those within +-1 hr of the ms data. Please proceed with caution...')
            col_ra = col_ra[idx]
            col_dec = col_dec[idx]
            col_mjd = col_mjd[idx]

        # Down-sample the ephemeris data to 60 s cadence (corresponding to <0.2" of shift due to solar rotation)
        dt_sec = np.median(np.diff(col_mjd)) * 24. * 3600.
        if dt_sec < 60.:
            if verbose:
                print('Ephemeris has a (median) cadence of {0:.3f}s. Down-sample to 60s.'.format(dt_sec))
            nt_res = int((col_mjd[-1] - col_mjd[0]) * 24. * 3600. / 60.)
            if nt_res > 2:
                btimes_res = np.linspace(col_mjd[0], col_mjd[-1], nt_res)
                f_ra = interp1d(col_mjd, col_ra, kind='linear')
                f_dec = interp1d(col_mjd, col_dec, kind='linear')
                ras_ = f_ra(np.array(btimes_res))
                decs_ = f_dec(np.array(btimes_res))
                ras = qa.convert(qa.quantity(ras_, 'deg'), 'rad')
                decs = qa.convert(qa.quantity(decs_, 'deg'), 'rad')
                btimes = btimes_res
        else:
            ras = qa.convert(qa.quantity(col_ra, 'deg'), 'rad')
            decs = qa.convert(qa.quantity(col_dec, 'deg'), 'rad')
            btimes = col_mjd

        ##TODO: I don't know why I was doing this. It seems it is never used.
        if interp_to_scan:
            f_ra = interp1d(col_mjd, col_ra, kind='linear')
            f_dec = interp1d(col_mjd, col_dec, kind='linear')
            ras_ = f_ra(np.array(btimes_scan))
            decs_ = f_dec(np.array(btimes_scan))
            ras = qa.convert(qa.quantity(ras_, 'deg'), 'rad')
            decs = qa.convert(qa.quantity(decs_, 'deg'), 'rad')
            btimes = btimes_scan
        msinfo['has_ephem_table'] = True
    else:
        if verbose:
            print('This measurement set does not have an ephemeris table attached. '
                  'The visibility phascenter uses RA and DEC of the FIELD attached to each scan')
        ras = ras_scan
        decs = decs_scan
        btimes = btimes_scan
        msinfo['has_ephem_table'] = False

    # put the relevent information into the dictionary
    btimestr = [qa.time(qa.quantity(btime, 'd'), form='fits', prec=10)[0] for btime in btimes]
    msinfo['vis'] = vis
    msinfo['observatory'] = observatory
    msinfo['scans'] = scans
    msinfo['fieldids'] = fieldids
    msinfo['inttimes'] = inttimes
    msinfo['scan_start_times'] = btimes_scan
    msinfo['scan_end_times'] = etimes_scan
    msinfo['btimes'] = btimes
    msinfo['btimestr'] = btimestr
    msinfo['ras'] = ras
    msinfo['decs'] = decs
    if msinfofile:
        np.savez(msinfofile, vis=vis, observatory=observatory, scans=scans, fieldids=fieldids, inttimes=inttimes,
                 btimes=btimes, btimestr=btimestr, ras=ras, decs=decs, has_ephem_table=msinfo['has_ephem_table'])
    return msinfo


def ephem_to_helio(vis=None, ephem=None, msinfo=None, reftime=None, dopolyfit=True,
                   usephacenter=True, geocentric=False, verbose=False):
    '''
    Module for calculating offsets of the phase center to the solar disk center using the following steps
    1. Take a solar ms database, read the scan and field information, find out the phase centers (in RA and DEC).
        This step is done with read_msinfo()
    2. Compare with the ephemeris of the solar disk center (in RA and DEC)
    3. Record RA/DEC of the phase center and offsets in RA/DEC and Helioprojective Cartesian coordinates (solar X/Y)
       inputs:
           msinfo: CASA MS information, output from read_msinfo
           ephem: solar ephem, output from read_horizons
           reftime: list of reference times (e.g., used for imaging)
                    CASA standard time format, either a single time (e.g., '2012/03/03/12:00:00'
                    or a time range (e.g., '2012/03/03/12:00:00~2012/03/03/13:00:00'. If the latter,
                    take the midpoint of the timerange for reference. If no date specified, take
                    the date of the first scan
           dopolyfit: Bool. Default: True. Works for MS database with only one source with continously tracking.
                    Disabled if usephacenter=False.
           usephacenter: Bool -- if True, correct for the RA and DEC in the ms file based on solar empheris.
                                 Otherwise assume the phasecenter is pointed to the solar disk center
                                 (EOVSA case)
           geocentric: Bool -- if True, use geocentric RA & DEC.
                        If False, use topocentric RA & DEC based on observatory location
                        Default: False

       return values:
           helio: a list of VLA pointing information
                   reftimestr: reference time, in FITS format string
                   reftime: reference time, in mjd format
                   ra: actual RA of phasecenter in the ms file at the reference time (interpolated)
                   dec: actual DEC of phasecenter in the ms file at the reference time (interpolated)
                   # CASA uses only RA and DEC of the closest field (e.g. in clean) #
                   ra_fld: RA of the CASA reference pointing direction, in radian
                   dec_fld: DEC of the CASA reference pointing direction, in radian
                   ra0: RA of the solar disk center, in radian
                   dec0: DEC of the solar disk center, in radian
                   raoff: RA offset of the phasecenter in the ms file to solar disk center, in arcsec
                   decoff: DEC offset of the phasecenter in the ms file to solar disk center, in arcsec
                   refx: heliocentric X offset of the phasecenter in the ms file to solar disk center, in arcsec
                   refy: heliocentric Y offset of the phasecenter in the ms file to solar disk center, in arcsec
                   has_ephem_table: flag to indicate if there is an ephemeris table attached.
    ######## Example #########
        from suncasa.utils import helioimage2fits as hf
        vis = vis = '22B-174_20221031_sun.1s.cal.ms'
        ephem = hf.ephem_to_helio(vis=vis, reftime='2022/10/31/20:37:10~2022/10/31/20:37:20', dopolyfit=True,
                        usephacenter=True, verbose=True)
        # Read out the ms information takes some time. To save time, one can read out the ms information first
            and supply the record here for registering multiple images. It will skip the read_msinfo() step.
        msinfo = hf.read_msinfo(vis=vis, verbose=True)
        ephem = hf.ephem_to_helio(vis=vis, msinfo=msinfo, reftime='2022/10/31/20:37:10~2022/10/31/20:37:20',
                                 dopolyfit=True, usephacenter=True, verbose=True)
    '''
    if not vis or not os.path.exists(vis):
        raise ValueError('Please provide information of the MS database!')

    if not msinfo:
        msinfo0 = read_msinfo(vis, verbose=verbose)
    else:
        if isinstance(msinfo, str):
            try:
                msinfo0 = np.load(msinfo)
            except:
                raise ValueError('The specified input msinfo file does not exist!')
        elif isinstance(msinfo, dict):
            msinfo0 = msinfo
        else:
            raise ValueError('msinfo should be either a numpy npz or a dictionary')
    if verbose:
        print('msinfo is derived from: {0:s}'.format(msinfo0['vis']))
    # scans = msinfo0['scans']
    fieldids = msinfo0['fieldids']
    btimes = msinfo0['btimes']
    btimestr = msinfo0['btimestr']
    inttimes = msinfo0['inttimes']
    ras = msinfo0['ras']
    decs = msinfo0['decs']
    scan_start_times = msinfo0['scan_start_times']
    if 'observatory' in msinfo0.keys():
        if msinfo0['observatory'] == 'EOVSA' or msinfo0['observatory'] == 'FASR':
            usephacenter = False

    if (not ephem) and usephacenter:
        if geocentric:
            # use geocentric ephemeris
            if verbose:
                print('Use geocentric RA & DEC')
            ephem = read_horizons(vis=vis, observatory='500')
        else:
            # use topocentric ephemeris
            if verbose:
                print('Use topocentric RA & DEC at observatory.')
            ephem = read_horizons(vis=vis, observatory=msinfo0['observatory'])

    if isinstance(ras, list):
        ra_rads = [ra['value'] for ra in ras]
    elif isinstance(ras, dict):
        ra_rads = ras['value']
    else:
        print('Type of msinfo0["ras"] unrecognized.')
        return 0
    if isinstance(decs, list):
        dec_rads = [dec['value'] for dec in decs]
    elif isinstance(decs, dict):
        dec_rads = decs['value']
    else:
        print('Type of msinfo0["decs"] unrecognized.')
    # fit 2nd order polynomial fits to the RAs and DECs #
    if usephacenter==False:
        dopolyfit = False

    if dopolyfit and (len(btimes) > 1):
        from suncasa.utils import fit_planet_position as fp
        if verbose:
            print('Use polynomial fit to interpolate the ms ephemeris data')
            print('Ephemeris start and end time', btimestr[0], btimestr[-1])
            print('Number of ephemeris data points used for polyfit', len(btimes))
        poly_res = fp.fit_planet_positions(btimes, ra_rads, dec_rads, allowed_error=0.05)
        if poly_res[0] == 0:  # success
            reftime_poly = poly_res[1]
            cra = np.flip(np.array(poly_res[2]))
            cdec = np.flip(np.array(poly_res[3]))
            ra_diff = (np.polyval(cra, btimes - reftime_poly) - ra_rads) * 180. / np.pi * 3600.
            dec_diff = (np.polyval(cdec, btimes - reftime_poly) - dec_rads) * 180. / np.pi * 3600.
            if verbose:
                print('Maximum deviation of the polynomial fit in RA, DEC is {0:.2f}", {1:.2f}"'.format(
                    np.nanmax(ra_diff), np.max(dec_diff)))
        else:
            print('Polyfit cannot achieve the required accuracy. Use linear interpolation instead.')

    # find out phase center infomation in ms according to the input time or timerange #
    if not reftime:
        raise ValueError('Please specify a reference time of the image')
    if isinstance(reftime, str):
        reftime = [reftime]
    if (not isinstance(reftime, list)):
        print('input "reftime" is not a valid list. Abort...')

    nreftime = len(reftime)
    helio = []
    for reftime0 in reftime:
        helio0 = dict.fromkeys(
            ['reftimestr', 'reftime', 'ra', 'dec', 'ra_fld', 'dec_fld', 'raoff', 'decoff', 'refx', 'refy', 'p0'])
        helio0['reftimestr'] = reftime0
        if '~' in reftime0:
            # if reftime0 is specified as a timerange
            try:
                [tbg0, tend0] = reftime0.split('~')
                tbg_d = qa.getvalue(qa.convert(qa.totime(tbg0), 'd'))[0]
                tend_d = qa.getvalue(qa.convert(qa.totime(tend0), 'd'))[0]
                tdur_s = (tend_d - tbg_d) * 3600. * 24.
                # if no date is specified, add up the date of the first scan
                if tend_d < 1.:
                    if tend_d >= tbg_d:
                        tend_d += int(btimes[0])
                    else:
                        tend_d += int(btimes[0]) + 1
                if tbg_d < 1.:
                    tbg_d += int(btimes[0])
                tref_d = (tbg_d + tend_d) / 2.
            except:
                print('Error in converting the input reftime: ' + str(reftime0) + '. Aborting...')
        else:
            # if reftime0 is specified as a single value
            try:
                tref_d = qa.getvalue(qa.convert(qa.totime(reftime0), 'd'))[0]
                # if no date is specified, add up the date of the first scan
                if tref_d < 1.:
                    tref_d += int(btimes[0])
                tbg_d = tref_d
                # use the intergration time
                # ind = bisect.bisect_left(btimes, tref_d)
                # if msinfo0['etimes']:
                #    tdur_s = inttimes[ind - 1]
                # else:
                #    tdur_s = np.mean(inttimes)
                tdur_s = 1.
            except:
                print('Error in converting the input reftime: ' + str(reftime0) + '. Aborting...')
        helio0['reftime'] = tref_d
        # helio0['date-obs'] = qa.time(qa.quantity(tbg_d, 'd'), form='fits', prec=10)[0]
        # helio0['exptime'] = tdur_s

        # find out phase center RA and DEC in the measurement set according to the reference time
        # if dopolyfit, then use the polynomial fit to interpolate
        inttime = np.nanmean(inttimes)
        ind = bisect.bisect_left(btimes, tref_d)
        if ind > 1:
            # meaning the reference time is AFTER the start time of the ms ephemeris data
            dt = tref_d - btimes[ind - 1]
            if ind < len(btimes):
                # the reference time falls into the time range of the ms ephemeris data
                scanlen = btimes[ind] - btimes[ind - 1]
                (ra_b, ra_e) = (ra_rads[ind - 1], ra_rads[ind])
                (dec_b, dec_e) = (dec_rads[ind - 1], dec_rads[ind])
            if ind >= len(btimes):
                # the reference time falls after the last ephemeris record
                print('Warning!!! The provided reference time falls AFTER the last ephemeris record.')
                print('            I have to use the last record to register the image')
                scanlen = btimes[ind - 1] - btimes[ind - 2]
                (ra_b, ra_e) = (ra_rads[ind - 2], ra_rads[ind - 1])
                (dec_b, dec_e) = (dec_rads[ind - 2], dec_rads[ind - 1])
        if ind == 1:  # only one recorded exists (e.g., data imported from AIPS)
            print('Warning!!! Only one record exists in the ms ephemeris record.')
            print('           I have to use this only record to register the image.')
            ra_b = ra_rads[ind - 1]
            ra_e = ra_b
            dec_b = dec_rads[ind - 1]
            dec_e = dec_b
            scanlen = 10.  # radom value
            dt = 0.
        if ind < 1:
            if np.abs((tref_d - btimes[0]) * 24 * 3600) < inttime / 2.0:
                if verbose:
                    print('The provided reference time {0:s} falls BEFORE the first ephemeris record {1:s}.'\
                            .format(Time(tref_d, format='mjd').isot, Time(btimes[0], format='mjd').isot))
                    print('However it is within 1/2 of the integration time {0:.1f}s w.r.t. the first record. I am ignoring this.'.format(inttime))
                ind = 1
                ra_b = ra_rads[ind - 1]
                ra_e = ra_b
                dec_b = dec_rads[ind - 1]
                dec_e = dec_b
                scanlen = 10.  # radom value
                dt = 0.
            else:
                raise ValueError('Reference time does not fall into the provided ephemeris record')
        if dopolyfit and (len(btimes) > 1) and poly_res[0] == 0:
            ra = np.polyval(cra, tref_d - reftime_poly)
            dec = np.polyval(cdec, tref_d - reftime_poly)
        else:
            ra = ra_b + (ra_e - ra_b) / scanlen * dt
            dec = dec_b + (dec_e - dec_b) / scanlen * dt
        if ra < 0:
            ra += 2. * np.pi
        if ra_b < 0:
            ra_b += 2. * np.pi

        # compare with ephemeris from JPL Horizons
        if not usephacenter:
            # Do not need to read the information from the measurement set
            if msinfo0['observatory'] == 'EVLA':
                observatory_id = '-5'
            elif msinfo0['observatory'] == 'EOVSA' or msinfo0['observatory'] == 'FASR' or msinfo0['observatory'] == 'OVRO_MMA':
                observatory_id = '-81'
            elif msinfo0['observatory'] == 'GMRT' or msinfo0['observatory'] == 'uGMRT':
                observatory_id = '399'
            elif msinfo0['observatory'] == 'ALMA':
                observatory_id = '-7'
            else:
                print('Observatory {} not recognized. Assume geocentric.'.format(msinfo0['observatory']))
                observatory_id = '500'

            if not ephem:
                ephem = read_horizons(Time(tref_d, format='mjd'), observatory=observatory_id)

        times_ephem = ephem['time']
        ras_ephem = ephem['ra']
        decs_ephem = ephem['dec']
        p0s_ephem = ephem['p0']

        if usephacenter:
            num_times_ephem=len(times_ephem)
            num_ras_ephem=len(ras_ephem)
            num_decs_ephem=len(decs_ephem)
            num_p0s_ephem=len(p0s_ephem)
            if len(times_ephem)>1 and num_times_ephem==num_ras_ephem\
                    and num_times_ephem==num_decs_ephem and num_times_ephem==num_p0s_ephem:
                f_ra = interp1d(times_ephem, ras_ephem, kind='linear')
                f_dec = interp1d(times_ephem, decs_ephem, kind='linear')
                f_p0 = interp1d(times_ephem, p0s_ephem, kind='linear')
                ra0 = f_ra([tref_d])[0]
                dec0 = f_dec([tref_d])[0]
                p0 = f_p0([tref_d])[0]
            else:
                try:
                    ra0 = ras_ephem[0]
                    dec0 = decs_ephem[0]
                    p0 = p0s_ephem[0]
                except:
                    print("Error in retrieving info from JPL solar ephemeris!")
            if ra0 < 0:
                ra0 += 2. * np.pi

            # RA and DEC offset in arcseconds
            decoff = degrees(normalize(dec - dec0)) * 3600.
            raoff = degrees(normalize(ra - ra0) * cos(dec)) * 3600.
            # Convert into heliocentric offsets
            prad = -radians(p0)
            refx = (-raoff) * cos(prad) - decoff * sin(prad)
            refy = (-raoff) * sin(prad) + decoff * cos(prad)
            helio0['raoff'] = raoff
            helio0['decoff'] = decoff
            helio0['refx'] = refx
            helio0['refy'] = refy
            helio0['ra0'] = ra0
            helio0['dec0'] = dec0
            helio0['p0'] = p0
        else:
            helio0['raoff'] = 0.
            helio0['decoff'] = 0.
            helio0['refx'] = 0.
            helio0['refy'] = 0.
            helio0['ra0'] = ras_ephem[0]
            helio0['dec0'] = decs_ephem[0]
            helio0['p0'] = p0s_ephem[0]

        helio0['ra'] = ra  # ra of the actual phase center
        helio0['dec'] = dec  # dec of the actual phase center
        ind = bisect.bisect_left(scan_start_times, tref_d)
        if ind > 1:
            dt = tref_d - scan_start_times[ind - 1]
            if ind < len(scan_start_times):
                fieldid = fieldids[ind]
                ms.open(vis)
                dir = ms.getfielddirmeas('PHASE_DIR', fieldid)
                ra_b = dir['m0']['value']
                dec_b = dir['m1']['value']
                if ra_b < 0:
                    ra_b += 2. * np.pi
                ms.close()
            if ind >= len(btimes):
                (ra_b, ra_e) = (ra_rads[ind - 2], ra_rads[ind - 1])
                (dec_b, dec_e) = (dec_rads[ind - 2], dec_rads[ind - 1])
        helio0['ra_fld'] = ra_b  # ra of the field, used as the reference in e.g., clean
        helio0['dec_fld'] = dec_b  # dec of the field, used as the refenrence in e.g., clean
        # helio['r_sun']=np.degrees(R_sun.value/(au.value*delta0))*3600. #in arcsecs
        ## add the tag whether an ephemeris table is attached or not
        helio0['has_ephem_table'] = msinfo0['has_ephem_table']
        helio.append(helio0)
    return helio


def getbeam(imagefile=None, beamfile=None):
    if not imagefile:
        raise ValueError('Please specify input images')
    bmaj = []
    bmin = []
    bpa = []
    beamunit = []
    bpaunit = []
    chans = []
    nimg = len(imagefile)
    for n in range(nimg):
        img = imagefile[n]
        if img is None:
            bmaj.append(None)
            bmin.append(None)
            bpa.append(None)
            beamunit.append(None)
            bpaunit.append(None)
            chans.append(None)
        bmaj_ = []
        bmin_ = []
        bpa_ = []
        if not os.path.exists(img):
            warnings.warn('{} does not exist!'.format(img))
            beamunit_ = []
            bpaunit_ = []
            chans_ = []
        else:
            ia.open(img)
            sum = ia.summary()
            ia.close()
            ia.done()
            if 'perplanebeams' in sum.keys():  # beam vary with frequency
                nbeams = sum['perplanebeams']['nChannels']
                beams = sum['perplanebeams']['beams']
                chans_ = [key[1:] for key in beams.keys()]
                chans_.sort(key=float)
                for chan in chans_:
                    bmaj0 = beams['*' + chan]['*0']['major']['value']
                    bmaj_.append(bmaj0)
                    bmin0 = beams['*' + chan]['*0']['minor']['value']
                    bmin_.append(bmin0)
                    bpa0 = beams['*' + chan]['*0']['positionangle']['value']
                    bpa_.append(bpa0)
                beamunit_ = beams['*' + chans_[0]]['*0']['major']['unit']
                bpaunit_ = beams['*' + chans_[0]]['*0']['positionangle']['unit']
            if 'restoringbeam' in sum.keys():  # only one beam
                bmaj_.append(sum['restoringbeam']['major']['value'])
                bmin_.append(sum['restoringbeam']['minor']['value'])
                bpa_.append(sum['restoringbeam']['positionangle']['value'])
                beamunit_ = sum['restoringbeam']['major']['unit']
                bpaunit_ = sum['restoringbeam']['positionangle']['unit']
                nbeams = 1
                chans_ = [0]

        bmaj.append(bmaj_)
        bmin.append(bmin_)
        bpa.append(bpa_)
        beamunit.append(beamunit_)
        bpaunit.append(bpaunit_)
        chans.append(chans_)
    if beamfile:  # write beams to ascii file
        print('Writing beam info to ascii file...')
        f = open(beamfile, 'w')
        f.write('CHANNEL No., BMAJ (' + beamunit[0] + '), BMIN (' + beamunit[0] + '), BPA (' + bpaunit[0] + ')')
        f.write("\n")
        for n in range(nimg):
            f.write('----For image: ' + imagefile[n] + '----')
            f.write('\n')
            chans_ = chans[n]
            for i in range(len(chans_)):
                f.write(str(chans_[i]) + ', ' + str(bmaj[n][i]) + ', ' + str(bmin[n][i]) + ', ' + str(bpa[n][i]))
                f.write("\n")
        f.close()
    return bmaj, bmin, bpa, beamunit, bpaunit


def imreg(vis=None, imagefile=None, timerange=None,
          ephem=None, msinfo=None, fitsfile=None,
          usephacenter=True, geocentric=False, dopolyfit=True, reftime=None, offsetfile=None, beamfile=None,
          toTb=False, sclfactor=1.0, p_ang=None, overwrite=True,
          deletehistory=False, subregion='', docompress=False, verbose=False):
    ''' 
    main routine to register CASA images
           Required Inputs:
               vis: STRING. CASA measurement set from which the image is derived
               imagefile: STRING or LIST. name of the input CASA image
               timerange: STRING or LIST. timerange used to generate the CASA image, must have the same length as the input images. 
                          Each element should be in CASA standard time format, e.g., '2012/03/03/12:00:00~2012/03/03/13:00:00'
           Optional Inputs:
               msinfo: DICTIONARY. CASA MS information, output from read_msinfo. If not provided, generate one from the supplied vis
               ephem: DICTIONARY. solar ephem, output from read_horizons. 
                      If not provided, query JPL Horizons based on time info of the vis (internet connection required)
               fitsfile: STRING or LIST. name of the output registered fits files
               toTb: Bool. Convert the default Jy/beam to brightness temperature?
               sclfactor: scale the image values up by its value (e.g., sclfactor = 100 to compensate VLA 20 dB attenuator)
               p_ang: solar p angle in degrees. If provided, use the supplied value and ignore the empheris
               verbose: Bool. Show more diagnostic info if True.
               usephacenter: Bool -- if True, correct for the RA and DEC in the ms file based on solar empheris.
                                     Otherwise assume the phasecenter is correctly pointed to the solar disk center
                                     (EOVSA case)
               geocentric: Bool -- if True, use geocentric RA & DEC.
                        If False, use topocentric RA & DEC based on observatory location
                        Default: False
               dopolyfit: Bool -- if True, fit the ephemeris from the measurement set using a polynomial fit.
                                  if False, just use linear interpolation
               ###### The following two parameters are only meant for temporary fixes ########################
               reftime: STRING or LIST. ONLY USED IF ANOTHER TIME (OTHER THAN TIME TO MAKE THE IMAGE)
                        IS NEEDED TO FIND RA AND DEC.
                        Each element should be in CASA standard time format, e.g., '2012/03/03/12:00:00'
               offsetfile: optionally provide an offset with a series of solar x and y offsets with timestamps
               ###############################################################################################
               subregion: only write the data within the sub-region selection. See 'help par.region' for details.
    Usage:
    >>> from suncasa.utils import helioimage2fits as hf
    >>> hf.imreg(vis='mydata.ms', imagefile='myimage.image', fitsfile='myimage.fits',
                 timerange='2017/08/21/20:21:10~2017/08/21/20:21:18')
    The output fits file is 'myimage.fits'

    History:
    BC (sometime in 2014): function was first wrote, followed by a number of edits by BC and SY
    BC (2019-07-16): Added checks for stokes parameter. Verified that for converting from Jy/beam to brightness temperature,
                     the convention of 2*k_b*T should always be used. I.e., for unpolarized source, stokes I, RR, LL, XX, YY, 
                     etc. in the output CASA images from (t)clean should all have same values of radio intensity 
                     (in Jy/beam) and brightness temperature (in K).

    '''

    if deletehistory:
        ms_clearhistory(vis)

    if not imagefile:
        raise ValueError('Please specify input image')
    if not timerange:
        raise ValueError('Please specify timerange of the input image')
    if isinstance(imagefile, str):
        imagefile = [imagefile]
    if isinstance(timerange, str):
        timerange = [timerange]
    if not fitsfile:
        fitsfile = []
        for img in imagefile:
            if img is None:
                fitsfile.append(None)
                continue
            if os.path.isdir(img):
                # input is CASA image
                fitsfile.append(img + '.fits')
            else:
                # input is not CASA image. Assuming FITS 
                if overwrite:
                    fitsfile.append(img)
                else:
                    fitsfile.append(img.replace('.fits', '.helio.fits'))

    if isinstance(fitsfile, str):
        fitsfile = [fitsfile]
    nimg = len(imagefile)
    if len(timerange) != nimg:
        raise ValueError('Number of input images does not equal to number of timeranges!')
    if len(fitsfile) != nimg:
        raise ValueError('Number of input images does not equal to number of output fits files!')
    nimg = len(imagefile)
    if verbose:
        print(str(nimg) + ' images to process...')

    if reftime:  # use as reference time to find solar disk RA and DEC to register the image, but not the actual timerange associated with the image
        if isinstance(reftime, str):
            reftime = [reftime] * nimg
        if len(reftime) != nimg:
            raise ValueError('Number of reference times does not match that of input images!')
        helio = ephem_to_helio(vis, ephem=ephem, msinfo=msinfo, reftime=reftime,
                               usephacenter=usephacenter, geocentric=geocentric, dopolyfit=dopolyfit, verbose=verbose)
    else:
        # use the supplied timerange to register the image
        helio = ephem_to_helio(vis, ephem=ephem, msinfo=msinfo, reftime=timerange,
                               usephacenter=usephacenter, geocentric=geocentric, dopolyfit=dopolyfit, verbose=verbose)

    if toTb:
        (bmajs, bmins, bpas, beamunits, bpaunits) = getbeam(imagefile=imagefile, beamfile=beamfile)

    for n, img in enumerate(imagefile):
        if verbose:
            print('processing image #' + str(n) + ' ' + img)
        if img is None:
            continue
        fitsf = fitsfile[n]
        timeran = timerange[n]
        # obtain duration of the image as FITS header exptime
        try:
            [tbg0, tend0] = timeran.split('~')
            tbg_d = qa.getvalue(qa.convert(qa.totime(tbg0), 'd'))[0]
            tend_d = qa.getvalue(qa.convert(qa.totime(tend0), 'd'))[0]
            tdur_s = (tend_d - tbg_d) * 3600. * 24.
            dateobs = qa.time(qa.quantity(tbg_d, 'd'), form='fits', prec=10)[0]
        except:
            print('Error in converting the input timerange: ' + str(timeran) + '. Proceeding to the next image...')
            continue

        hel = helio[n]
        if not (p_ang is None):
            # if p angle is provided as a keyword, overwrite the one from ephemeris
            try:
                p_ang = normalize(p_ang)
                hel['p0'] = p_ang
            except ValueError:
                print('Provided p angle value is invalid. Use p angle from ephemeris.')
        if not os.path.exists(img):
            warnings.warn('{} does not existed!'.format(img))
        else:
            if os.path.exists(fitsf) and not overwrite:
                print('Specified fits file already exists and overwrite is set to False. I will do nothing...')
                continue
            else:
                # check if the input image is already helioimage2fits registered
                if not os.path.isdir(img):
                    hdu = pyfits.open(img)
                    if 'hel_reg' in hdu[0].header:
                        if hdu[0].header['hel_reg']:
                            print('The input image has already been processed by helioimage2fits.py. I will do nothing..')
                        else:
                            print('The input image has already been processed by helioimage2fits.py, but somehow unsuccessful.')
                            print('Please use the original image and redo it...')
                        continue
                    else:
                        if verbose:
                            print('Input image is probably a fits file. Will try to proceed.')
                    hdu.close()
                p0 = hel['p0']
                if os.path.exists(img + '/logtable'):
                    tb.open(img + '/logtable', nomodify=False)
                    nobs = tb.nrows()
                    tb.removerows([i + 1 for i in range(nobs - 1)])
                    tb.close()
                ia.open(img)
                imr = ia.rotate(pa=str(-p0) + 'deg')
                if subregion != '':
                    imr = imr.subimage(region=subregion)
                imr.tofits(fitsf, history=False, overwrite=overwrite)
                imr.close()
                imsum = ia.summary()
                ia.close()
                ia.done()

            # construct the standard fits header
            # RA and DEC of the reference pixel crpix1 and crpix2
            (imra, imdec) = (imsum['refval'][0], imsum['refval'][1])
            if imra < 0:
                imra += 2. * np.pi
            if verbose:
                print('image ra', imra)
                print('visibility ra', hel['ra'])
                print('field ra', hel['ra_fld'])
                print('image dec', imdec)
                print('visibility dec', hel['dec'])
                print('field dec', hel['dec_fld'])

            ## When (t)clean is making an image, the default center of the image is coordinates of the associated FIELD
            ## If a new "phasecenter" is supplied in (t)clean, if
            #       CASE A (<2014-ish): No ephemeris table is attached. The visibility phase center
            #                              is the same as the RA and DEC of the FIELD
            #       CASA B (>2014-ish): Ephemeris table is attached. The visibility phase center
            #                              is the same as that from the ephemeris, and is
            #                              different from the RA and DEC of the FIELD.
            #### find out the difference between the image phase center and RA and DEC of the associated FIELD in arrsec
            ddec_fld = degrees(normalize(imdec - hel['dec_fld'])) * 3600.
            dra_fld = degrees(normalize(imra - hel['ra_fld']) * cos(hel['dec_fld'])) * 3600.

            # Convert into image heliocentric offsets
            prad = -radians(hel['p0'])
            dx_fld = (-dra_fld) * cos(prad) - ddec_fld * sin(prad)
            dy_fld = (-dra_fld) * sin(prad) + ddec_fld * cos(prad)

            #### find out the difference between the image phase center and RA and DEC of the visibility phasecenter
            ddec_vis = degrees(normalize(imdec - hel['dec'])) * 3600.  # in arcsec
            dra_vis = degrees(normalize(imra - hel['ra']) * cos(hel['dec'])) * 3600.  # in arcsec
            # Convert into image heliocentric offsets
            dx_vis = (-dra_vis) * cos(prad) - ddec_vis * sin(prad)
            dy_vis = (-dra_vis) * sin(prad) + ddec_vis * cos(prad)
            if offsetfile:
                try:
                    offset = np.load(offsetfile)
                except:
                    raise ValueError('The specified offsetfile does not exist!')
                reftimes_d = offset['reftimes_d']
                xoffs = offset['xoffs']
                yoffs = offset['yoffs']
                timg_d = hel['reftime']
                ind = bisect.bisect_left(reftimes_d, timg_d)
                xoff = xoffs[ind - 1]
                yoff = yoffs[ind - 1]
            else:
                xoff = hel['refx']
                yoff = hel['refy']
            if verbose:
                print('offset of image phase center to FIELD phase center in RA and DEC (arcsec): '
                      'dx={0:.2f}, dy={1:.2f}'.format(dra_fld, ddec_fld))
                print('offset of image phase center to FIELD phase center in helioprojective arcsec: '
                      'dx={0:.2f}, dy={1:.2f}'.format(dx_fld, dy_fld))
                print('offset of image phase center to VISIBILITY phase center in helioprojective arcsec: '
                      'dx={0:.2f}, dy={1:.2f}'.format(dx_vis, dy_vis))
                print('offset of VISIBILITY phase center to solar disk center in helioprojective arcsec: '
                      'dx={0:.2f}, dy={1:.2f}'.format(xoff, yoff))
            if hel['has_ephem_table']:
                if verbose:
                    print('This ms has an ephemeris table attached. Use ephemeris phase center as reference')
                (crval1, crval2) = (xoff + dx_vis, yoff + dy_vis)
            else:
                if verbose:
                    print('This ms does not have an ephemeris table attached. Use FIELD phase center as reference.')
                (crval1, crval2) = (xoff + dx_fld, yoff + dy_fld)
            # update the fits header to heliocentric coordinates

            hdu = pyfits.open(fitsf, mode='update')
            hdu[0].verify('fix')
            header = hdu[0].header
            dshape = hdu[0].data.shape
            ndim = hdu[0].data.ndim
            (cdelt1, cdelt2) = (
                -header['cdelt1'] * 3600., header['cdelt2'] * 3600.)  # Original CDELT1, 2 are for RA and DEC in degrees
            header['cdelt1'] = cdelt1
            header['cdelt2'] = cdelt2
            header['cunit1'] = 'arcsec'
            header['cunit2'] = 'arcsec'
            header['crval1'] = crval1
            header['crval2'] = crval2
            header['ctype1'] = 'HPLN-TAN'
            header['ctype2'] = 'HPLT-TAN'
            header['date-obs'] = dateobs  # begin time of the image
            if tdur_s:
                exptime = tdur_s
            else:
                exptime = 1.
            p_angle = hel['p0']
            hgln_obs = 0.
            rsun_ref = sun.constants.radius.value
            if sunpy1:
                dsun_obs = sun.earth_distance(Time(dateobs)).to(u.meter).value
                rsun_obs = sun.angular_radius(Time(dateobs)).value
                hglt_obs = sun.B0(Time(dateobs)).value
            else:
                dsun_obs = sun.sunearth_distance(Time(dateobs)).to(u.meter).value
                rsun_obs = sun.solar_semidiameter_angular_size(Time(dateobs)).value
                hglt_obs = sun.heliographic_solar_center(Time(dateobs))[1].value
            try:
                # this works for pyfits version of CASA 4.7.0 but not CASA 4.6.0
                header.set('exptime', exptime)
                header.set('p_angle', p_angle)
                header.set('dsun_obs', dsun_obs)
                header.set('rsun_obs', rsun_obs)
                header.set('rsun_ref', rsun_ref)
                header.set('hgln_obs', hgln_obs)
                header.set('hglt_obs', hglt_obs)
            except:
                # this works for astropy.io.fits
                header.append(('exptime', exptime))
                header.append(('p_angle', p_angle))
                header.append(('dsun_obs', dsun_obs))
                header.append(('rsun_obs', rsun_obs))
                header.append(('rsun_ref', rsun_ref))
                header.append(('hgln_obs', hgln_obs))
                header.append(('hglt_obs', hglt_obs))

            # check if stokes parameter exist
            exist_stokes = False
            stokes_mapper = {1: 'I', 2: 'Q', 3: 'U', 4: 'V', -1: 'RR', -2: 'LL',
                             -3: 'RL', -4: 'LR', -5: 'XX', -6: 'YY', -7: 'XY', -8: 'YX'}
            if 'CRVAL3' in header.keys():
                if header['CTYPE3'] == 'STOKES':
                    stokenum = header['CRVAL3']
                    exist_stokes = True
            if 'CRVAL4' in header.keys():
                if header['CTYPE4'] == 'STOKES':
                    stokenum = header['CRVAL4']
                    exist_stokes = True
            if exist_stokes:
                if stokenum in stokes_mapper.keys():
                    stokesstr = stokes_mapper[stokenum]
                else:
                    print('Stokes parameter {0:d} not recognized. Assuming Stokes I'.format(stokenum))
                    stokenum = 1
                    stokesstr = 'I'
                if verbose:
                    print('This image is in Stokes ' + stokesstr)
            else:
                print('STOKES Information does not seem to exist! Assuming Stokes I')
                stokenum = 1

            keys = list(header.keys())
            values = list(header.values())
            # which axis is frequency?
            faxis = keys[values.index('FREQ')][-1]
            faxis_ind = ndim - int(faxis)
            # insert the frequency axis information into the header as they may be stripped off if FITS compressed.
            header['REFFRQ'] = header['CRVAL' + faxis]
            header['DLTFRQ'] = header['CDELT' + faxis]
            header['FQUNIT'] = header['CUNIT' + faxis]

            data = hdu[0].data  # remember the data order is reversed due to the FITS convension
            # intensity units to brightness temperature
            if toTb:
                # get restoring beam info
                bmaj = bmajs[n]
                bmin = bmins[n]
                beamunit = beamunits[n]
                bpa = bpas[n]
                bpaunit = bpaunits[n]
                # find out the polarization of this image
                k_b = qa.constants('k')['value']
                c_l = qa.constants('c')['value']
                # Always use 2*kb for all polarizations
                const = 2. * k_b / c_l ** 2
                bpatmp = qa.quantity(bpa, bpaunit)['value'] - qa.convert(qa.quantity(p0, 'deg'), bpaunit)['value']
                header['BPA'] = bpatmp[0]
                if header['BUNIT'].lower() == 'jy/beam':
                    header['BUNIT'] = 'K'
                    header['BTYPE'] = 'Brightness Temperature'
                    for i in range(dshape[faxis_ind]):
                        nu = header['CRVAL' + faxis] + header['CDELT' + faxis] * (i + 1 - header['CRPIX' + faxis])
                        if header['CUNIT' + faxis] == 'KHz':
                            nu *= 1e3
                        if header['CUNIT' + faxis] == 'MHz':
                            nu *= 1e6
                        if header['CUNIT' + faxis] == 'GHz':
                            nu *= 1e9
                        if len(bmaj) > 1:  # multiple (per-plane) beams
                            bmajtmp = bmaj[i]
                            bmintmp = bmin[i]
                        else:  # one single beam
                            bmajtmp = bmaj[0]
                            bmintmp = bmin[0]
                        if beamunit == 'arcsec':
                            bmaj0 = np.radians(bmajtmp / 3600.)
                            bmin0 = np.radians(bmintmp / 3600.)
                        if beamunit == 'arcmin':
                            bmaj0 = np.radians(bmajtmp / 60.)
                            bmin0 = np.radians(bmintmp / 60.)
                        if beamunit == 'deg':
                            bmaj0 = np.radians(bmajtmp)
                            bmin0 = np.radians(bmintmp)
                        if beamunit == 'rad':
                            bmaj0 = bmajtmp
                            bmin0 = bmintmp
                        beam_area = bmaj0 * bmin0 * np.pi / (4. * log(2.))
                        factor = const * nu ** 2  # SI unit
                        jy_to_si = 1e-26
                        # print(nu/1e9, beam_area, factor)
                        if faxis == '3':
                            data[:, i, :, :] *= jy_to_si / beam_area / factor
                        if faxis == '4':
                            data[i, :, :, :] *= jy_to_si / beam_area / factor

            try:
                header.append(('hel_reg', True))
                header.append(('history', 'Converted by helioimage2fits.py'))
            except:
                header.set('hel_reg', True)
                header.set('history', 'Converted by helioimage2fits.py')

            data *= sclfactor
            header = ndfits.headerfix(header)
            hdu.flush()
            hdu.close()

            if ndim - np.count_nonzero(np.array(dshape) == 1) > 3:
                docompress = False
                '''
                    Caveat: only 1D, 2D, or 3D images are currently supported by
                    the astropy fits compression. If a n-dimensional image data array
                    does not have at least n-3 single-dimensional entries,
                    force docompress to be False
                '''

                print(
                    'warning: The fits data contains more than 3 non squeezable dimensions. Skipping fits compression..')
            if docompress:
                fitsftmp = fitsf + ".tmp.fits"
                os.system("mv {} {}".format(fitsf, fitsftmp))
                hdu = pyfits.open(fitsftmp)
                hdu[0].verify('fix')
                header = hdu[0].header
                data = hdu[0].data
                ndfits.write(fitsf, data, header, compression_type='RICE_1',
                             quantize_level=4.0)
                os.system("rm -rf {}".format(fitsftmp))
    if deletehistory:
        ms_restorehistory(vis)
    return fitsfile


def calc_phasecenter_from_solxy(vis, timerange='', xycen=None, usemsphacenter=True, observatory=None):
    '''
    return the phase center in RA and DEC of a given solar coordinates

    :param vis: input measurement sets file
    :param timerange: can be a string or astropy.time.core.Time object, or a 2-element list of string or Time object
    :param xycen:  solar x-pos and y-pos in arcsec
    :param usemsphacenter:
    :return:
    phasecenter
    midtim: mid time of the given timerange
    '''
    ms.open(vis)
    metadata = ms.metadata()
    if not observatory:
        observatory = metadata.observatorynames()[0]
        if metadata.observatorynames()[0] == 'EVLA':
            observatory = '-5'
        elif metadata.observatorynames()[0] == 'EOVSA' or metadata.observatorynames()[0] == 'FASR' or metadata.observatorynames()[0] == 'OVRO_MMA':
            observatory = '-81'
        elif metadata.observatorynames()[0] == 'GMRT' or metadata.observatorynames()[0] == 'uGMRT':
            observatory_id = '399'
        elif metadata.observatorynames()[0] == 'ALMA':
            observatory = '-7'
       
    try:
        mstrange = metadata.timerangeforobs(0)
        tst = Time(mstrange['begin']['m0']['value'], format='mjd')
        ted = Time(mstrange['end']['m0']['value'], format='mjd')
    except:
        print('Something is wrong in using metadata tool. Maybe you are using an old CASA version.')
        print('Try to use the summary outputs.')
        out = ms.summary()
        tst = Time(out['BeginTime'], format='mjd')
        ted = Time(out['EndTime'], format='mjd')
    ms.close()
    metadata.close()
    datstr = tst.iso[:10]

    if isinstance(timerange, Time):
        try:
            (sttim, edtim) = timerange
        except:
            sttim = timerange
            edtim = sttim
    else:
        if timerange == '':
            sttim = tst
            edtim = ted
        else:
            try:
                (tstart, tend) = timerange.split('~')
                if tstart[2] == ':':
                    sttim = Time(datstr + 'T' + tstart)
                    edtim = Time(datstr + 'T' + tend)
                    # timerange = '{0}/{1}~{0}/{2}'.format(datstr.replace('-', '/'), tstart, tend)
                else:
                    sttim = Time(qa.quantity(tstart, 'd')['value'], format='mjd')
                    edtim = Time(qa.quantity(tend, 'd')['value'], format='mjd')
            except:
                try:
                    if timerange[2] == ':':
                        sttim = Time(datstr + 'T' + timerange)
                        edtim = sttim
                    else:
                        sttim = Time(qa.quantity(timerange, 'd')['value'], format='mjd')
                        edtim = sttim
                except ValueError:
                    print("keyword 'timerange' in wrong format")

    midtim_mjd = (sttim.mjd + edtim.mjd) / 2.
    midtim = Time(midtim_mjd, format='mjd')
    eph = read_horizons(t0=midtim,observatory=observatory)
    if observatory == 'EOVSA' or (not usemsphacenter):
        print('Phasecenter in the ms is ignored')
        # use RA and DEC from FIELD ID 0
        tb.open(vis + '/FIELD')
        phadir = tb.getcol('PHASE_DIR').flatten()
        tb.close()
        ra0 = phadir[0]
        dec0 = phadir[1]
    else:
        ra0 = eph['ra'][0]
        dec0 = eph['dec'][0]

    if not xycen:
        # use solar disk center as default
        phasecenter = 'J2000 ' + str(ra0) + 'rad ' + str(dec0) + 'rad'
    else:
        x0 = np.radians(xycen[0] / 3600.)
        y0 = np.radians(xycen[1] / 3600.)
        p0 = np.radians(eph['p0'][0])  # p angle in radians
        raoff = -((x0) * np.cos(p0) - y0 * np.sin(p0)) / np.cos(eph['dec'][0])
        decoff = (x0) * np.sin(p0) + y0 * np.cos(p0)
        newra = ra0 + raoff
        newdec = dec0 + decoff
        phasecenter = 'J2000 ' + str(newra) + 'rad ' + str(newdec) + 'rad'
    return phasecenter, midtim


from astropy.coordinates import SkyCoord, EarthLocation
from astropy import units as u
from astropy.io import fits
import sunpy.map
from sunpy.coordinates import frames, sun

def radec_fits_to_helio(fits_in, helio_sunpy_fits_name =None, obs_loc = EarthLocation(lat=37.232259*u.deg, lon=-118.28479*u.deg)):
    """
        composing a sunpy compatible fits from a fits file with RA and DEC in the header
        using location of OVSA as example


        :param fits_in: input fits file
        :param helio_sunpy_fits_name: output fits file name, if None, the output file will be named as fits_in.replace('.fits', '_heliosunpy.fits')
        :param obs_loc: location of the observer, default is OVSA
    """

    hdu = fits.open(fits_in)
    header = hdu[0].header
    obstime = Time(header['date-obs'])

    data = hdu[0].data[0, 0, :, :]
    frequency = header['crval3']*u.Hz
    obs_gcrs = SkyCoord(obs_loc.get_gcrs(obstime))
    reference_coord = SkyCoord(header['crval1']*u.Unit(header['cunit1']),
            header['crval2']*u.Unit(header['cunit2']),frame='gcrs',obstime=obstime,
            obsgeoloc=obs_gcrs.cartesian, obsgeovel=obs_gcrs.velocity.to_cartesian(),
            distance=obs_gcrs.hcrs.distance)
    reference_coord_arcsec = reference_coord.transform_to(frames.Helioprojective(observer=obs_gcrs))
    cdelt1 = (np.abs(header['cdelt1'])*u.deg).to(u.arcsec)
    cdelt2 = (np.abs(header['cdelt2'])*u.deg).to(u.arcsec)
    P1 = sun.P(obstime)
    new_header = sunpy.map.make_fitswcs_header(data, reference_coord_arcsec,
        reference_pixel=u.Quantity([header['crpix1']-1, header['crpix2']-1]*u.pixel),
        scale=u.Quantity([cdelt1, cdelt2]*u.arcsec/u.pix), wavelength=3e8/frequency.to(u.Hz).value*u.m,
        rotation_angle=-P1, observatory='obs')
    obsview_map = sunpy.map.Map(data, new_header)
    fov_x = new_header["naxis1"]*cdelt1
    fov_y = new_header["naxis2"]*cdelt2

    obsview_map_rotate = obsview_map.rotate()
    bl = SkyCoord(-fov_x/2, -fov_y/2, frame=obsview_map_rotate.coordinate_frame)
    tr = SkyCoord(fov_x/2,  fov_y/2,  frame=obsview_map_rotate.coordinate_frame)
    obsview_submap = obsview_map_rotate.submap(bl, top_right=tr)

    if helio_sunpy_fits_name is None:
        helio_sunpy_fits_name = fits_in.replace('.fits', '_heliosunpy.fits')
    obsview_submap.save(helio_sunpy_fits_name, overwrite=True)


from astropy.io import fits
import astropy.table
from astropy.time import Time
import matplotlib.pyplot as plt
import matplotlib.colors as colors
from matplotlib.dates import DateFormatter
import numpy as np
from mpl_toolkits.axes_grid1 import make_axes_locatable


def get_dspec(filename, doplot=False, vmax=None, vmin=None, norm=None, cmap=None):
    """
    Reads and optionally plots an EOVSA Dynamic Spectrum from a FITS file.

    This function reads a FITS file containing EOVSA dynamic spectrum data, returning the data as a dictionary. If requested, it also generates a plot of the dynamic spectrum.

    Parameters
    ----------
    filename : str
        Path and name of the EOVSA dynamic spectrum FITS file to read.
    doplot : bool, optional
        If True, generates a plot of the dynamic spectrum. Default is False.
    vmin : float, optional
        Minimum value for the color scale. Ignored if `norm` is provided. Default is None, which autoscales.
    vmax : float, optional
        Maximum value for the color scale. Ignored if `norm` is provided. Default is None, which autoscales.
    norm : `matplotlib.colors.Normalize` or None, optional
        Normalization for the color scale of the plot. If None, a linear scaling is used. Default is None.
    cmap : str or `matplotlib.colors.Colormap`, optional
        Colormap for the dynamic spectrum plot. Can be a colormap name or an instance. Default is None, which uses the default colormap.

    Returns
    -------
    dict
        A dictionary containing the dynamic spectrum data (`spectrogram`), frequency axis (`spectrum_axis` in GHz), and time axis (`time_axis` in modified Julian date).

    Example
    -------
    ::

        from suncasa.eovsa import eovsa_dspec as ds
        from astropy.time import Time
        from matplotlib.colors import LogNorm

        # Read EOVSA Dynamic Spectrum FITS file <filename>
        filename = 'EOVSA_TPall_20170713.fts'
        s = ds.get_dspec(filename, doplot=True, cmap='gist_heat', norm=LogNorm(vmax=2.1e3, vmin=40))

        # To access the data in the spectrogram object, use
        spec = s['spectrogram']  # Array of amplitudes in SFU, of size nfreq,ntimes
        fghz = s['spectrum_axis']  # Array of frequencies in GHz, of size nfreq
        tim = Time(s['time_axis'], format='mjd')   # Array of UT times in astropy.time object, of size ntimes


    """

    hdulist = fits.open(filename)
    spec = hdulist[0].data
    fghz = np.array(astropy.table.Table(hdulist[1].data)['sfreq'])
    tim = astropy.table.Table(hdulist[2].data)
    tmjd= np.array(tim['mjd']) + np.array(tim['time']) / 24. / 3600 / 1000
    tim = Time(tmjd, format='mjd')
    timplt = tim.plot_date
    ntim = len(timplt)
    nfreq = len(fghz)
    if doplot:
        fig, ax = plt.subplots(figsize=(7, 4))
        # if vmax is None:
        #     vmax = np.nanmax(spec)
        # if vmin is None:
        #     vmin = 0.1
        # if norm is None:
        #     norm = colors.Normalize(vmax=vmax, vmin=vmin)
        pcm = ax.pcolormesh(timplt, fghz, spec, norm=norm, vmax=vmax, vmin=vmin, cmap=cmap,rasterized='True')
        ax.xaxis_date()
        ax.xaxis.set_major_formatter(DateFormatter("%H:%M"))
        ax.set_ylim(fghz[0], fghz[-1])
        ax.set_xlim(tim[0].plot_date, tim[-1].plot_date)
        ax.set_xlabel('Time [UT]')
        ax.set_ylabel('Frequency [GHz]')
        ax.set_title('EOVSA Dynamic Spectrum for ' + tim[0].datetime.strftime('%Y-%m-%d'))
        divider = make_axes_locatable(ax)
        cax_spec = divider.append_axes('right', size='1.5%', pad=0.05)
        clb_spec = plt.colorbar(pcm, ax=ax, cax=cax_spec, label='Flux density [sfu]')
        cmap = pcm.get_cmap()
        bkg_facecolor=cmap(0.0)
        ax.set_facecolor(bkg_facecolor)
        dtim = np.diff(tmjd)
        idxs_tgap = np.where(dtim*24*60 >= 0.5)[0]
        if len(idxs_tgap)>0:
            for idx in idxs_tgap:
                if idx+1 < len(tmjd):
                    ax.axvspan(tim[idx].plot_date,tim[idx+1].plot_date,facecolor=bkg_facecolor,edgecolor='none')



        def format_coord(x, y):
            col = np.argmin(np.absolute(timplt - x))
            row = np.argmin(np.absolute(fghz - y))
            if col >= 0 and col < ntim and row >= 0 and row < nfreq:
                timstr = tim[col].isot
                flux = spec[row, col]
                return 'time {0}, freq = {1:.3f} GHz, flux = {2:.2f} sfu'.format(timstr, y, flux)
            else:
                return 'x = {0}, y = {1:.3f}'.format(x, y)

        ax.format_coord = format_coord
        fig.tight_layout()
        plt.show()
    return {'spectrogram': spec, 'spectrum_axis': fghz, 'time_axis': tmjd}



##################### generated by xml-casa (v2) from pmaxfit.xml ###################
##################### b3be0c8ff5a66315d1c40770a10aa871 ##############################
from __future__ import absolute_import
from casashell.private.stack_manip import find_local as __sf__
from casashell.private.stack_manip import find_frame as _find_frame
from casatools.typecheck import validator as _pc
from casatools.coercetype import coerce as _coerce
from suncasatasks import pmaxfit as _pmaxfit_t
from collections import OrderedDict
import numpy
import sys
import os

import shutil

def static_var(varname, value):
    def decorate(func):
        setattr(func, varname, value)
        return func
    return decorate

class _pmaxfit:
    """
    pmaxfit ---- Find maximum and do parabolic fit in the sky

    
    PARAMETER SUMMARY
    imagename        Name of the input image
    box              Rectangular region(s) to select in direction plane. See "help par.box"
    for details. Default is to use the entire direction plane.
    eg "100, 120, 200, 220, 300, 300, 400, 400" to use two boxes.
    
    OVERVIEW
    This application finds the pixel with the maximum value in the region, and then uses function
    findsources to generate a Componentlist with one component.
    
    The method returns a dictionary with fours keys, 'succeeded', 'timestamps', 'imagenames'
    and 'outputs'. The value of 'outputs' is a dictionary representing
    a component list reflecting the fit results over multiple channels.
    Both the 'outputs' dictionaries can be read into a component list tool (default tool is named cl)
    using the fromrecord() method for easier inspection using tool methods, eg
    
    FITTING OVER MULTIPLE CHANNELS
    
    For fitting over multiple channels, the result of the previous successful fit is used as
    the estimate for the next channel. The number of gaussians fit cannot be varied on a channel
    by channel basis. Thus the variation of source structure should be reasonably smooth in
    frequency to produce reliable fit results.
    

    --------- parameter descriptions ---------------------------------------------

    imagefiles A list of the input images
    ncpu       Number of cpu cores to use
    box        Rectangular region(s) to select in direction plane. See "help par.box" for details. Default is to use the entire direction plane.
    width      Half-width of fit grid
    [1;42mRETURNS[1;m       void

    --------- examples -----------------------------------------------------------

    
    
    
    EXAMPLE:
    
    Here is how one might fit two gaussians to multiple channels of a cube using the fit
    from the previous channel as the initial estimate for the next. It also illustrates
    how one can specify a region in the associated continuum image as the region to use
    as the fit for the channel.
    
    
    begin{verbatim}
    default pmaxfit
    imagename = "co_cube.im"
    # specify region using region from continuum
    box = "100,120,200,200"
    pmaxfit()
    


    """

    _info_group_ = """analysis"""
    _info_desc_ = """Find maximum and do parabolic fit in the sky"""

    __schema = {'imagefiles': {'type': 'cReqPathVec', 'coerce': [_coerce.to_list,_coerce.expand_pathvec]}, 'ncpu': {'type': 'cInt'}, 'box': {'type': 'cStr', 'coerce': _coerce.to_str}, 'width': {'type': 'cInt'}}

    def __init__(self):
        self.__stdout = None
        self.__stderr = None
        self.__root_frame_ = None

    def __globals_(self):
        if self.__root_frame_ is None:
            self.__root_frame_ = _find_frame( )
            assert self.__root_frame_ is not None, "could not find CASAshell global frame"
        return self.__root_frame_

    def __to_string_(self,value):
        if type(value) is str:
            return "'%s'" % value
        else:
            return str(value)

    def __validate_(self,doc,schema):
        return _pc.validate(doc,schema)

    def __do_inp_output(self,param_prefix,description_str,formatting_chars):
        out = self.__stdout or sys.stdout
        description = description_str.split( )
        prefix_width = 23 + 10 + 4
        output = [ ]
        addon = ''
        first_addon = True
        while len(description) > 0:
            ## starting a new line.....................................................................
            if len(output) == 0:
                ## for first line add parameter information............................................
                if len(param_prefix)-formatting_chars > prefix_width - 1:
                    output.append(param_prefix)
                    continue
                addon = param_prefix + ' #'
                first_addon = True
                addon_formatting = formatting_chars
            else:
                ## for subsequent lines space over prefix width........................................
                addon = (' ' * prefix_width) + '#'
                first_addon = False
                addon_formatting = 0
            ## if first word of description puts us over the screen width, bail........................
            if len(addon + description[0]) - addon_formatting + 1 > self.term_width:
                ## if we're doing the first line make sure it's output.................................
                if first_addon: output.append(addon)
                break
            while len(description) > 0:
                ## if the next description word puts us over break for the next line...................
                if len(addon + description[0]) - addon_formatting + 1 > self.term_width: break
                addon = addon + ' ' + description[0]
                description.pop(0)
            output.append(addon)
        out.write('\n'.join(output) + '\n')

    #--------- return nonsubparam values ----------------------------------------------

    def __imagefiles_dflt( self, glb ):
        return [  ]

    def __imagefiles( self, glb ):
        if 'imagefiles' in glb: return glb['imagefiles']
        return [  ]

    def __ncpu_dflt( self, glb ):
        return int(8)

    def __ncpu( self, glb ):
        if 'ncpu' in glb: return glb['ncpu']
        return int(8)

    def __box_dflt( self, glb ):
        return ''

    def __box( self, glb ):
        if 'box' in glb: return glb['box']
        return ''

    def __width_dflt( self, glb ):
        return int(5)

    def __width( self, glb ):
        if 'width' in glb: return glb['width']
        return int(5)



    #--------- return inp/go default --------------------------------------------------


    #--------- return subparam values -------------------------------------------------


    #--------- subparam inp output ----------------------------------------------------
    def __imagefiles_inp(self):
        description = ''
        value = self.__imagefiles( self.__globals_( ) )
        (pre,post) = ('','') if self.__validate_({'imagefiles': value},{'imagefiles': self.__schema['imagefiles']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-10.10s = %s%-23s%s' % ('imagefiles',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __ncpu_inp(self):
        description = ''
        value = self.__ncpu( self.__globals_( ) )
        (pre,post) = ('','') if self.__validate_({'ncpu': value},{'ncpu': self.__schema['ncpu']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-10.10s = %s%-23s%s' % ('ncpu',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __box_inp(self):
        description = ''
        value = self.__box( self.__globals_( ) )
        (pre,post) = ('','') if self.__validate_({'box': value},{'box': self.__schema['box']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-10.10s = %s%-23s%s' % ('box',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))
    def __width_inp(self):
        description = ''
        value = self.__width( self.__globals_( ) )
        (pre,post) = ('','') if self.__validate_({'width': value},{'width': self.__schema['width']}) else ('\x1B[91m','\x1B[0m')
        self.__do_inp_output('%-10.10s = %s%-23s%s' % ('width',pre,self.__to_string_(value),post),description,0+len(pre)+len(post))

    #--------- global default implementation-------------------------------------------
    @static_var('state', __sf__('casa_inp_go_state'))
    def set_global_defaults(self):
        self.set_global_defaults.state['last'] = self
        glb = self.__globals_( )
        if 'imagefiles' in glb: del glb['imagefiles']
        if 'ncpu' in glb: del glb['ncpu']
        if 'box' in glb: del glb['box']
        if 'width' in glb: del glb['width']


    #--------- inp function -----------------------------------------------------------
    def inp(self):
        print("# pmaxfit -- %s" % self._info_desc_)
        self.term_width, self.term_height = shutil.get_terminal_size(fallback=(80, 24))
        self.__imagefiles_inp( )
        self.__ncpu_inp( )
        self.__box_inp( )
        self.__width_inp( )

    #--------- tget function ----------------------------------------------------------
    @static_var('state', __sf__('casa_inp_go_state'))
    def tget(self,file=None):
        from casashell.private.stack_manip import find_frame
        from runpy import run_path
        filename = None
        if file is None:
            if os.path.isfile("pmaxfit.last"):
                filename = "pmaxfit.last"
        elif isinstance(file, str):
            if os.path.isfile(file):
                filename = file
        if filename is not None:
            glob = find_frame( )
            newglob = run_path( filename, init_globals={ } )
            for i in newglob:
                glob[i] = newglob[i]
            self.tget.state['last'] = self
        else:
            print("could not find last file, setting defaults instead...")
            self.set_global_defaults( )

    def __call__( self, imagefiles=None, ncpu=None, box=None, width=None ):
        def noobj(s):
           if s.startswith('<') and s.endswith('>'):
               return "None"
           else:
               return s
        _prefile = os.path.realpath('pmaxfit.pre')
        _postfile = os.path.realpath('pmaxfit.last')
        _return_result_ = None
        _arguments = [imagefiles,ncpu,box,width]
        _invocation_parameters = OrderedDict( )
        if any(map(lambda x: x is not None,_arguments)):
            # invoke python style
            # set the non sub-parameters that are not None
            local_global = { }
            if imagefiles is not None: local_global['imagefiles'] = imagefiles
            if ncpu is not None: local_global['ncpu'] = ncpu
            if box is not None: local_global['box'] = box
            if width is not None: local_global['width'] = width

            # the invocation parameters for the non-subparameters can now be set - this picks up those defaults
            _invocation_parameters['imagefiles'] = self.__imagefiles( local_global )
            _invocation_parameters['ncpu'] = self.__ncpu( local_global )
            _invocation_parameters['box'] = self.__box( local_global )
            _invocation_parameters['width'] = self.__width( local_global )

            # the sub-parameters can then be set. Use the supplied value if not None, else the function, which gets the appropriate default
            

        else:
            # invoke with inp/go semantics
            _invocation_parameters['imagefiles'] = self.__imagefiles( self.__globals_( ) )
            _invocation_parameters['ncpu'] = self.__ncpu( self.__globals_( ) )
            _invocation_parameters['box'] = self.__box( self.__globals_( ) )
            _invocation_parameters['width'] = self.__width( self.__globals_( ) )
        try:
            with open(_prefile,'w') as _f:
                for _i in _invocation_parameters:
                    _f.write("%-10s = %s\n" % (_i,noobj(repr(_invocation_parameters[_i]))))
                _f.write("#pmaxfit( ")
                count = 0
                for _i in _invocation_parameters:
                    _f.write("%s=%s" % (_i,noobj(repr(_invocation_parameters[_i]))))
                    count += 1
                    if count < len(_invocation_parameters): _f.write(",")
                _f.write(" )\n")
        except: pass
        try:
            _return_result_ = _pmaxfit_t( _invocation_parameters['imagefiles'],_invocation_parameters['ncpu'],_invocation_parameters['box'],_invocation_parameters['width'] )
        except Exception as e:
            from traceback import format_exc
            from casatasks import casalog
            casalog.origin('pmaxfit')
            casalog.post("Exception Reported: Error in pmaxfit: %s" % str(e),'SEVERE')
            casalog.post(format_exc( ))
            _return_result_ = False
        try:
            os.rename(_prefile,_postfile)
        except: pass
        return _return_result_

pmaxfit = _pmaxfit( )




#
# Based on function called in VLA observing script /home/mchost/evla/include/fit_planet_position.py
#
# Original author bjb @ nrao in summer 2012
#
# python functions to fit planet ephemerides and return polynomials
# for ra, dec, and distance.
#
# Slightly modified by Bin Chen @ NJIT on Dec 27, 2022

from math import fabs, sqrt, pow
import numpy as np
#
# this is the singular value decomposition stuff.  it should come from
# numpy, but jython doesn't include it, so i just went ahead and did it
# by hand.  if jython ever includes numpy, then do the following:
#
# from numpy import *
# from scipy import array
#
# as an aside, i really only need searchsorted, polyfit, poly1d, and 
# tolist from numpy, but i can't figure out how to import tolist without 
# importing the whole damn module ><.
#
# then, in the code below, look for the comments related to numpy and
# make the fixes there as well.  also, you don't need the bisect import
# here, because searchsorted does it for you.
#

import bisect


def fpoly(x, np):
    p = [1.0]
    for ii in range(1, np):
        p.append(p[ii - 1] * x)
    return p


def svdfit(xx, yy, sig, ma):
    aa = []
    bb = []
    for ii in range(len(xx)):
        afunc = fpoly(xx[ii], ma)
        tmp = 1.0 / sig[ii]
        taa = []
        for jj in range(ma):
            taa.append(afunc[jj] * tmp)
        bb.append(yy[ii] * tmp)
        aa.append(taa)
    (uu, ww, vv) = svd(aa)
    wmax = max(ww)
    thresh = 2.0e-12 * wmax
    for ii in range(ma):
        if ww[ii] < thresh:
            ww[ii] = 0.0
    aa = svbksb(uu, ww, vv, bb)
    chisq = 0.0
    for ii in range(len(xx)):
        afunc = fpoly(xx[ii], ma)
        model = 0.0
        for jj in range(ma):
            model += aa[jj] * afunc[jj]
        chisq += ((yy[ii] - model) / sig[ii]) * ((yy[ii] - model) / sig[ii])
    return (aa, uu, vv, ww, chisq)


# Almost exact translation of the ALGOL SVD algorithm published in
# Numer. Math. 14, 403-420 (1970) by G. H. Golub and C. Reinsch
#
# Copyright (c) 2005 by Thomas R. Metcalf, helicity314-stitch <at> yahoo <dot> com
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# Pure Python SVD algorithm.
# Input: 2-D list (m by n) with m >= n
# Output: U,W V so that A = U*W*VT
#    Note this program returns V not VT (=transpose(V))
#    On error, a ValueError is raised.
#
# Here is the test case (first example) from Golub and Reinsch
#
# a = [[22.,10., 2.,  3., 7.],
#      [14., 7.,10.,  0., 8.],
#      [-1.,13.,-1.,-11., 3.],
#      [-3.,-2.,13., -2., 4.],
#      [ 9., 8., 1., -2., 4.],
#      [ 9., 1.,-7.,  5.,-1.],
#      [ 2.,-6., 6.,  5., 1.],
#      [ 4., 5., 0., -2., 2.]]
#
# import svd
# import math
# u,w,v = svd.svd(a)
# print w
#
# [35.327043465311384, 1.2982256062667619e-15,
#  19.999999999999996, 19.595917942265423, 0.0]
#
# the correct answer is (the order may vary)
#
# print (math.sqrt(1248.),20.,math.sqrt(384.),0.,0.)
#
# (35.327043465311391, 20.0, 19.595917942265423, 0.0, 0.0)
#
# transpose and matrix multiplication functions are also included
# to facilitate the solution of linear systems.
#
# Version 1.0 2005 May 01

import copy
import math


def svd(a):
    '''Compute the singular value decomposition of array.'''

    # Golub and Reinsch state that eps should not be smaller than the
    # machine precision, ie the smallest number
    # for which 1+e>1.  tol should be beta/e where beta is the smallest
    # positive number representable in the computer.
    eps = 1.e-15  # assumes double precision
    tol = 1.e-64 / eps
    assert 1.0 + eps > 1.0  # if this fails, make eps bigger
    assert tol > 0.0  # if this fails, make tol bigger
    itmax = 50
    u = copy.deepcopy(a)
    m = len(a)
    n = len(a[0])
    #   if __debug__: print 'a is ',m,' by ',n

    if m < n:
        if __debug__: print('Error: m is less than n')
        raise ValueError('SVD Error: m is less than n.')

    e = [0.0] * n  # allocate arrays
    q = [0.0] * n
    v = []
    for k in range(n): v.append([0.0] * n)

    # Householder's reduction to bidiagonal form

    g = 0.0
    x = 0.0

    for i in range(n):
        e[i] = g
        s = 0.0
        l = i + 1
        for j in range(i, m): s += (u[j][i] * u[j][i])
        if s <= tol:
            g = 0.0
        else:
            f = u[i][i]
            if f < 0.0:
                g = math.sqrt(s)
            else:
                g = -math.sqrt(s)
            h = f * g - s
            u[i][i] = f - g
            for j in range(l, n):
                s = 0.0
                for k in range(i, m): s += u[k][i] * u[k][j]
                f = s / h
                for k in range(i, m): u[k][j] = u[k][j] + f * u[k][i]
        q[i] = g
        s = 0.0
        for j in range(l, n): s = s + u[i][j] * u[i][j]
        if s <= tol:
            g = 0.0
        else:
            f = u[i][i + 1]
            if f < 0.0:
                g = math.sqrt(s)
            else:
                g = -math.sqrt(s)
            h = f * g - s
            u[i][i + 1] = f - g
            for j in range(l, n): e[j] = u[i][j] / h
            for j in range(l, m):
                s = 0.0
                for k in range(l, n): s = s + (u[j][k] * u[i][k])
                for k in range(l, n): u[j][k] = u[j][k] + (s * e[k])
        y = abs(q[i]) + abs(e[i])
        if y > x: x = y
    # accumulation of right hand gtransformations
    for i in range(n - 1, -1, -1):
        if g != 0.0:
            h = g * u[i][i + 1]
            for j in range(l, n): v[j][i] = u[i][j] / h
            for j in range(l, n):
                s = 0.0
                for k in range(l, n): s += (u[i][k] * v[k][j])
                for k in range(l, n): v[k][j] += (s * v[k][i])
        for j in range(l, n):
            v[i][j] = 0.0
            v[j][i] = 0.0
        v[i][i] = 1.0
        g = e[i]
        l = i
    # accumulation of left hand transformations
    for i in range(n - 1, -1, -1):
        l = i + 1
        g = q[i]
        for j in range(l, n): u[i][j] = 0.0
        if g != 0.0:
            h = u[i][i] * g
            for j in range(l, n):
                s = 0.0
                for k in range(l, m): s += (u[k][i] * u[k][j])
                f = s / h
                for k in range(i, m): u[k][j] += (f * u[k][i])
            for j in range(i, m): u[j][i] = u[j][i] / g
        else:
            for j in range(i, m): u[j][i] = 0.0
        u[i][i] += 1.0
    # diagonalization of the bidiagonal form
    eps = eps * x
    for k in range(n - 1, -1, -1):
        for iteration in range(itmax):
            # test f splitting
            for l in range(k, -1, -1):
                goto_test_f_convergence = False
                if abs(e[l]) <= eps:
                    # goto test f convergence
                    goto_test_f_convergence = True
                    break  # break out of l loop
                if abs(q[l - 1]) <= eps:
                    # goto cancellation
                    break  # break out of l loop
            if not goto_test_f_convergence:
                # cancellation of e[l] if l>0
                c = 0.0
                s = 1.0
                l1 = l - 1
                for i in range(l, k + 1):
                    f = s * e[i]
                    e[i] = c * e[i]
                    if abs(f) <= eps:
                        # goto test f convergence
                        break
                    g = q[i]
                    h = pythag(f, g)
                    q[i] = h
                    c = g / h
                    s = -f / h
                    for j in range(m):
                        y = u[j][l1]
                        z = u[j][i]
                        u[j][l1] = y * c + z * s
                        u[j][i] = -y * s + z * c
            # test f convergence
            z = q[k]
            if l == k:
                # convergence
                if z < 0.0:
                    # q[k] is made non-negative
                    q[k] = -z
                    for j in range(n):
                        v[j][k] = -v[j][k]
                break  # break out of iteration loop and move on to next k value
            if iteration >= itmax - 1:
                if __debug__: print('Error: no convergence.')
                # should this move on the the next k or exit with error??
                # raise ValueError,'SVD Error: No convergence.'  # exit the program with error
                break  # break out of iteration loop and move on to next k
            # shift from bottom 2x2 minor
            x = q[l]
            y = q[k - 1]
            g = e[k - 1]
            h = e[k]
            f = ((y - z) * (y + z) + (g - h) * (g + h)) / (2.0 * h * y)
            g = pythag(f, 1.0)
            if f < 0:
                f = ((x - z) * (x + z) + h * (y / (f - g) - h)) / x
            else:
                f = ((x - z) * (x + z) + h * (y / (f + g) - h)) / x
            # next QR transformation
            c = 1.0
            s = 1.0
            for i in range(l + 1, k + 1):
                g = e[i]
                y = q[i]
                h = s * g
                g = c * g
                z = pythag(f, h)
                e[i - 1] = z
                c = f / z
                s = h / z
                f = x * c + g * s
                g = -x * s + g * c
                h = y * s
                y = y * c
                for j in range(n):
                    x = v[j][i - 1]
                    z = v[j][i]
                    v[j][i - 1] = x * c + z * s
                    v[j][i] = -x * s + z * c
                z = pythag(f, h)
                q[i - 1] = z
                c = f / z
                s = h / z
                f = c * g + s * y
                x = -s * g + c * y
                for j in range(m):
                    y = u[j][i - 1]
                    z = u[j][i]
                    u[j][i - 1] = y * c + z * s
                    u[j][i] = -y * s + z * c
            e[l] = 0.0
            e[k] = f
            q[k] = x
            # goto test f splitting

    # vt = transpose(v)
    # return (u,q,vt)
    return (u, q, v)


def pythag(a, b):
    absa = abs(a)
    absb = abs(b)
    if absa > absb:
        return absa * math.sqrt(1.0 + (absb / absa) ** 2)
    else:
        if absb == 0.0:
            return 0.0
        else:
            return absb * math.sqrt(1.0 + (absa / absb) ** 2)


def svbksb(uu, ww, vv, bb):
    ma = len(uu[0])
    nn = len(uu)
    tmp = []
    for ii in range(ma):
        ss = 0.0
        if ww[ii] != 0.0:
            for jj in range(nn):
                ss += uu[jj][ii] * bb[jj]
            ss /= ww[ii]
        tmp.append(ss)
    xx = []
    for ii in range(ma):
        ss = 0.0
        for jj in range(ma):
            ss += vv[ii][jj] * tmp[jj]
        xx.append(ss)
    return xx


def svdvar(vv, ww, ma):
    wti = []
    for ii in range(ma):
        wti.append(0.0)
        if ww[ii] != 0.0:
            wti[ii] = 1.0 / (ww[ii] * ww[ii])
    cvm = [[0 for col in range(ma)] for row in range(ma)]
    for ii in range(ma):
        for jj in range(ii + 1):
            sum = 0.0
            for kk in range(ma):
                sum += vv[ii][kk] * vv[jj][kk] * wti[kk]
            cvm[ii][jj] = sum
            cvm[jj][ii] = sum
    return cvm


def fit_planet_positions(times, ras, decs, start_time=None, end_time=None, distances=None, allowed_error=0.01):
    '''
    find a fitting polynomial for an ephemeris table.

    inputs:
        times = list of times at which the ephemeris positions are
                tabulated.  assumed sorted in ascending order.
                (MJD).
        ras = list of right ascensions at those times (radians).
        decs = list of declinations at those times (radians).
    Optional inputs:
        start_time = the start time of the SB (MJD).
        end_time = the end time of the SB (MJD).
        distances = list of distances at those times (AU).
        allowed_error = the allowed error in the fitting polynomials
                        for ra and dec from the tabulated values
                        (asec).

    returned is a list, first element is the return status:
        0 -> success
        1-7 -> Warning: did not converge to required accuracy.  
               1 -> right ascension only didn't converge
               2 -> declination only didn't converge
               3 -> right ascension and declination didn't converge
               4 -> distance only didn't converge
               5 -> right ascension and distance didn't converge
               6 -> declination and distance didn't converge
               7 -> none of the three converged
               (note that in this case the best fitted polynomials 
               are still returned [N.B. i should return the error 
               somewhere...].)
        8 -> Error: the time range from start_time to end_time is
             not completely contained in the tabulated times.
    second element is the time t0 to which the polynomial is
        referenced.
    third element is the list of right ascension coefficients.
    fourth element is the list of declination coefficients.
    fifth element is the list of distance coefficients.

    bjb
    nrao
    summer 2012
    '''

    NPOLYMAX = 7
    ASEC2RAD = 2.0626480624710e5

    times = np.array(times)
    ras = np.array(ras)
    decs = np.array(decs)

    #
    # check the times
    #
    if start_time or end_time:
        if start_time < times[0] or end_time > times[-1]:
            result = [8]
            return result

    result = [0]

    #
    # convert from arcseconds to radians
    #
    allowed_error /= ASEC2RAD

    #
    # carve out the sub-lists for the requested time range
    #
    # if numpy ever gets into jython, use the following
    # instead of what is below...
    #
    #   start_index = searchsorted(times, start_time)
    #   end_index = searchsorted(times, end_time)
    #
    # use the bisect functions in lieu of numpy/searchsorted
    #
    if start_time:
        start_index = bisect.bisect_left(times, start_time)
        start_index = max(start_index - 3, 0)
    else:
        start_index = 0
    if end_time:
        end_index = bisect.bisect_right(times, end_time)
    else:
        end_index = len(times) - 1

    end_index = min(end_index + 3, len(times) - 1)
    times_slice = times[start_index:end_index]
    ras_slice = ras[start_index:end_index]
    decs_slice = decs[start_index:end_index]

    if distances:
        distances_slice = distances[start_index:end_index]

    #
    # reference times to the center time, otherwise the
    # polynomials have huge coefficients and potentially
    # blow up
    #
    t0 = times_slice[int(len(times_slice) / 2)]
    #print('length of the time array used for polyfit', len(times_slice))
    result.append(t0)
    times_slice -= t0
    #for ii in range(len(times_slice)):
    #    times_slice[ii] -= t0

    #
    # make the corresponding scipy/numpy array quantities
    #
    #   time_array = array(times_slice)
    #   ra_array = array(ras_slice)
    #   dec_array = array(decs_slice)
    #   distance_array = array(distances_slice)
    time_array = times_slice
    ra_array = ras_slice
    dec_array = decs_slice
    if distances:
        distance_array = distances_slice

    #
    # right ascension fit
    #
    #
    # for non-numpy version, need errors.  just take them all equal.
    #
    sig = []
    for ii in range(len(ra_array)):
        sig.append(ra_array[0] / 1000.0)
    max_ra_difn = allowed_error + 1.0
    npoly_ra = 2
    while max_ra_difn > allowed_error and npoly_ra <= NPOLYMAX:
        #
        # if numpy ever gets into jython...
        #
        #       a_ra = polyfit (time_array, ra_array, npoly_ra)
        #       ra_fit = poly1d(a_ra)
        #
        # polyfit returns the coefficients in reverse order from normal,
        # so reverse them back...
        #
        #       a_ra = a_ra[::-1]
        #       max_ra_difn = 0.0
        #       for ii in range(len(time_array)):
        #           model_ra = ra_fit(time_array[ii])
        #           max_ra_difn = max (max_ra_difn, fabs(model_ra-ra_array[ii]))
        #
        # this is the numpy replacement, done by hand...
        #
        (a_ra, uu, vt, ww, chisq) = svdfit(time_array, ra_array, sig, npoly_ra)
        #
        # i never use the covariance matrix, so could comment this out,
        # but i leave it in, in case i ever want to check on the significance
        # of the coefficients (compare them against their error) to see whether
        # there is a problem.
        #
        cvm = svdvar(vt, ww, npoly_ra)
        max_ra_difn = 0.0
        for ii in range(len(time_array)):
            model_ra = 0.0
            for jj in range(npoly_ra):
                model_ra += a_ra[jj] * pow(time_array[ii], jj)
            max_ra_difn = max(max_ra_difn, fabs(model_ra - ra_array[ii]))
        npoly_ra += 1
    npoly_ra -= 1
    if npoly_ra == NPOLYMAX and max_ra_difn > allowed_error:
        result[0] = 1
    #
    # the numpy version - have to convert from numpy array to list
    #
    #   result.append(a_ra.tolist())
    result.append(a_ra)

    #
    # declination fit
    #
    sig = []
    for ii in range(len(dec_array)):
        sig.append(dec_array[0] / 1000.0)
    max_dec_difn = allowed_error + 1.0
    npoly_dec = 1
    while max_dec_difn > allowed_error and npoly_dec <= NPOLYMAX:
        # numpy
        #       a_dec = polyfit (time_array, dec_array, npoly_dec)
        #       dec_fit = poly1d(a_dec)
        #       a_dec = a_dec[::-1]
        #       max_dec_difn = 0.0
        #       for ii in range(len(time_array)):
        #           model_dec = dec_fit(time_array[ii])
        #           max_dec_difn = max (max_dec_difn, fabs(model_dec-dec_array[ii]))
        (a_dec, uu, vt, ww, chisq) = svdfit(time_array, dec_array, sig, npoly_dec)
        cvm = svdvar(vt, ww, npoly_dec)
        max_dec_difn = 0.0
        for ii in range(len(time_array)):
            model_dec = 0.0
            for jj in range(npoly_dec):
                model_dec += a_dec[jj] * pow(time_array[ii], jj)
            max_dec_difn = max(max_dec_difn, fabs(model_dec - dec_array[ii]))
        npoly_dec += 1
    npoly_dec -= 1
    if npoly_dec == NPOLYMAX and max_dec_difn > allowed_error:
        result[0] += 2
    #   result.append(a_dec.tolist())
    result.append(a_dec)

    #
    # distance fit.  for distance, set the allowed error to ~1.5 km...
    #
    if distances:
        sig = []
        for ii in range(len(distance_array)):
            sig.append(distance_array[0] / 1000.0)
        allowed_error = 1.0e-8
        max_distance_difn = allowed_error + 1.0
        npoly_distance = 1
        while max_distance_difn > allowed_error and npoly_distance <= NPOLYMAX:
            # numpy
            #       a_distance = polyfit (time_array, distance_array, npoly_distance)
            #       distance_fit = poly1d(a_distance)
            #       a_distance = a_distance[::-1]
            #       max_distance_difn = 0.0
            #       for ii in range(len(time_array)):
            #           model_distance = distance_fit(time_array[ii])
            #           max_distance_difn = max (max_distance_difn, fabs(model_distance-distance_array[ii]))
            (a_distance, uu, vt, ww, chisq) = svdfit(time_array, distance_array, sig, npoly_distance)
            cvm = svdvar(vt, ww, npoly_distance)
            max_distance_difn = 0.0
            for ii in range(len(time_array)):
                model_distance = 0.0
                for jj in range(npoly_distance):
                    model_distance += a_distance[jj] * pow(time_array[ii], jj)
                max_distance_difn = max(max_distance_difn, fabs(model_distance - distance_array[ii]))
            npoly_distance += 1
        npoly_distance -= 1
        if npoly_distance == NPOLYMAX and max_distance_difn > allowed_error:
            result[0] += 4
        #   result.append(a_distance.tolist())
        result.append(a_distance)

    return result



import numpy as np
from astropy.time import Time
import matplotlib.pyplot as plt
import matplotlib.colors as colors
import matplotlib.colorbar as colorbar
from suncasa.utils import fitsutils as fu
from sunpy import map as smap
import copy

## SDO/AIA image fits
aiafile = 'aia.lev1_euv_12s.2021-05-07T190209Z.171.image_lev1.fits'
# aiafile = 'aia.lev1_uv_24s.2021-05-07T190214Z.1600.image_lev1.fits'
## EOVSA image fits created above
eofile = 'EOVSA_20210507T190205.000000.outim.image.allbd.fits'

## image center ("xycen"), pixel scale ("cell"), and image field of view ("fov")
## for the plots
xycen = np.array([-895, 290])
fov = np.array([200, 200])
xlim = xycen[0] + np.array([-1, 1]) * 0.5 * fov[0]
ylim = xycen[1] + np.array([-1, 1]) * 0.5 * fov[1]
## Select the lower 48 spectral windows for plotting
fidxs = np.arange(0, 48)
## The alpha blending value for , between 0 (transparent) and 1 (opaque).
alpha = 0.5
## The respective maximum intensity of EOVSA images
vmin = 0.3 


## Get solar coordinates of every pixels in the radio image.
rmap, ndim, npol_fits, stokaxis, rcfreqs, rdata, rheader = fu.read_compressed_image_fits(eofile)
nspw = len(rcfreqs)
eodate = Time(rmap.date.mjd + rmap.exposure_time.value / 2. / 24 / 3600, format='mjd')
ny, nx = rmap.data.shape
x0, x1 = (np.array([1, rmap.meta['NAXIS1']]) - rmap.meta['CRPIX1']) * rmap.meta['CDELT1'] + \
         rmap.meta['CRVAL1']
y0, y1 = (np.array([1, rmap.meta['NAXIS2']]) - rmap.meta['CRPIX2']) * rmap.meta['CDELT2'] + \
         rmap.meta['CRVAL2']
dx = rmap.meta['CDELT1']
dy = rmap.meta['CDELT2']
mapx, mapy = np.linspace(x0, x1, nx), np.linspace(y0, y1, ny)

## Plot the AIA maps on the background
fig, axs = plt.subplots(nrows=1, ncols=2, figsize=(10, 4.8), sharex=True, sharey=True)
aiacmap = copy.copy(plt.get_cmap('gray_r'))
aiacmap.set_under('white')
aiacmap.set_over('k')
aiamap = smap.Map(aiafile)
ax = axs[0]
aiamap.plot(axes=ax, cmap=aiacmap)
ax = axs[1]
aiamap.plot(axes=ax, cmap=aiacmap)

## Plot EOVSA images as filled contour on top of the AIA image
vmins = np.array([vmin] * len(rcfreqs))
spwcmap = plt.get_cmap('RdYlBu')
colors_spws = spwcmap(np.linspace(0, 1, nspw))
cts = []

for idx, fidx in enumerate(fidxs):
    data = rdata[0, fidx, ...]
    clevels1 = np.linspace(vmins[idx], 1.0, 2) * np.nanmax(data)
    cts.append(ax.contourf(mapx, mapy, data, levels=clevels1,
                           # colors=[colors_spws[fidx]] * len(clevels1), ## contour
                           colors=[colors_spws[fidx]],  ## contourf
                           alpha=alpha))
ax.set_xlim(xlim)
ax.set_ylim(ylim)
for ax in axs:
    ax.set_xlabel('Solar-X [arcsec]')
    ax.set_ylabel('Solar-y [arcsec]')
    ax.set_title('')
ax.text(0.02, 0.98, 'SOL2021-05-07 M4 Event', ha='left', va='top', color='k', transform=ax.transAxes, fontsize=12,
        fontweight='bold')
ax.text(0.02, 0.01, ' '.join(['AIA {:.0f} Å'.format(aiamap.wavelength.value), aiamap.date.datetime.strftime('%Y-%m-%dT%H:%M:%S')]), ha='left',
        va='bottom',
        color='k', transform=ax.transAxes)
ax.text(0.02, 0.05, ' '.join(['EOVSA     ', eodate.datetime.strftime('%Y-%m-%dT%H:%M:%S')]), ha='left', va='bottom',
        color='k', transform=ax.transAxes)

## Add the color bar for the EOVSA multi-frequency images
ax1_pos = ax.get_position().extents
caxwidth = (ax1_pos[2] - ax1_pos[0]) * 0.1
caxcenter = ax1_pos[2] + caxwidth * 0.75
cayheight = ax1_pos[3] - 0.06 - ax1_pos[1]
rcfreqsplt = rcfreqs
bdwidth = np.nanmean(np.diff(rcfreqsplt))
bounds = np.linspace(rcfreqsplt[0] - bdwidth / 2.0, rcfreqsplt[-1] + bdwidth / 2.0, len(rcfreqsplt) + 1)
ticks = rcfreqsplt[1:][::3]
cax = plt.axes((caxcenter - caxwidth / 2.0, ax1_pos[1], caxwidth, cayheight))

cb = colorbar.ColorbarBase(cax, norm=colors.Normalize(vmax=rcfreqsplt[-1], vmin=rcfreqsplt[0]), cmap=plt.cm.RdYlBu,
                           orientation='vertical', boundaries=bounds, spacing='uniform', ticks=ticks,
                           format='%4.1f  ', alpha=1)
plt.text(0.5, 1.05, 'MW', ha='center', va='bottom', transform=cax.transAxes, color='k', fontweight='normal')
plt.text(0.5, 1.01, '[GHz]', ha='center', va='bottom', transform=cax.transAxes, color='k', fontweight='normal')

cax.xaxis.set_visible(False)
cax.tick_params(axis="y", direction="in", pad=-21.5, length=0, colors='k', labelsize=8)

fig.tight_layout()



import os
import matplotlib.pyplot as plt
import matplotlib.colors as colors
from sunpy import map as smap
import astropy.units as u
from mpl_toolkits.axes_grid1 import make_axes_locatable
import matplotlib.colorbar as colorbar
import matplotlib.patches as patches
from datetime import timedelta
from datetime import datetime
from glob import glob
import numpy as np
from astropy.time import Time
import calendar
from suncasa.utils import plot_mapX as pmX
import urllib.request
from sunpy.physics.differential_rotation import diffrot_map
from suncasa.utils import DButil
from tqdm import tqdm
from suncasa.eovsa import eovsa_readfits as er

# imgfitsdir = '/Users/fisher/myworkspace/'
# imgfitstmpdir = '/Users/fisher/myworkspace/fitstmp/'
# pltfigdir = '/Users/fisher/myworkspace/SynopticImg/eovsamedia/eovsa-browser/'

imgfitsdir = '/data1/eovsa/fits/synoptic/'
imgfitstmpdir = '/data1/workdir/fitstmp/'
pltfigdir = '/common/webplots/SynopticImg/eovsamedia/eovsa-browser/'


def clearImage():
    for (dirpath, dirnames, filenames) in os.walk(pltfigdir):
        for filename in filenames:
            for k in ['0094', '0193', '0335', '4500', '0171', '0304', '0131', '1700', '0211', '1600', '_HMIcont',
                      '_HMImag']:
                # for k in ['_Halph_fr']:
                if k in filename:
                    print(os.path.join(dirpath, filename))
                    os.system('rm -rf ' + os.path.join(dirpath, filename))


def pltEovsaQlookImageSeries(timobjs, spws, vmaxs, vmins, aiawave, bd, fig=None, axs=None, imgoutdir=None, overwrite=False,
                             verbose=False):
    from astropy.visualization.stretch import AsinhStretch
    from astropy.visualization import ImageNormalize
    plt.ioff()
    imgfiles = []
    dpi = 512. / 4
    aiaDataSource = {"0094": 8,
                     "0193": 11,
                     "0335": 14,
                     # "4500": 17,
                     "0171": 10,
                     "0304": 13,
                     "0131": 9,
                     "1700": 16,
                     "0211": 12,
                     # "1600": 15,
                     "_HMIcont": 18,
                     "_HMImag": 19}

    tmjd = timobjs.mjd
    tmjd_base = np.floor(tmjd)
    tmjd_hr = (tmjd - tmjd_base) * 24
    for tidx, timobj in enumerate(tqdm(timobjs)):
        dateobj = timobj.to_datetime()
        timestr = dateobj.strftime("%Y-%m-%dT%H:%M:%SZ")
        tstrname = dateobj.strftime("%Y%m%dT%H%M%SZ")
        datestrdir = dateobj.strftime("%Y/%m/%d/")
        imgindir = imgfitsdir + datestrdir
        timobj_prevday = Time(tmjd[tidx] - 1, format='mjd')
        dateobj_prevday = timobj_prevday.to_datetime()
        datestrdir_prevday = dateobj_prevday.strftime("%Y/%m/%d/")
        imgindir_prevday = imgfitsdir + datestrdir_prevday

        # if not os.path.exists(imgindir): os.makedirs(imgindir)
        cmap = plt.get_cmap('sdoaia304')

        if verbose: print('Processing EOVSA images for date {}'.format(dateobj.strftime('%Y-%m-%d')))
        key, sourceid = aiawave, aiaDataSource[aiawave]
        s, sp = bd, spws[bd]

        figoutname = os.path.join(imgoutdir, 'eovsa_bd{:02d}_aia{}_{}.jpg'.format(s + 1, key, tstrname))

        if overwrite or (not os.path.exists(figoutname)):
            ax = axs[0]
            ax.cla()
            spwstr = '-'.join(['{:02d}'.format(int(sp_)) for sp_ in sp.split('~')])
            t_hr = tmjd_hr[tidx]
            t_hr_st_blend = 2.0
            t_hr_ed_blend = 14.0
            if t_hr <= 8.0:
                eofile = imgindir_prevday + 'eovsa_{}.spw{}.tb.disk.fits'.format(dateobj_prevday.strftime('%Y%m%d'),
                                                                                 spwstr)
            else:
                eofile = imgindir + 'eovsa_{}.spw{}.tb.disk.fits'.format(dateobj.strftime('%Y%m%d'), spwstr)
            if not os.path.exists(eofile):
                continue

            stretch = AsinhStretch(a=0.15)
            norm = ImageNormalize(vmin=vmins[s], vmax=vmaxs[s], stretch=stretch)
            eomap = er.readfits(eofile)
            eomap = eomap.resample(u.Quantity(eomap.dimensions) / 2)
            eomap.data[np.isnan(eomap.data)] = 0.0
            eomap_rot = diffrot_map(eomap, time=timobj)
            offlimbidx = np.where(eomap_rot.data == eomap_rot.data[0, 0])
            eomap_rot.data[offlimbidx] = eomap.data[offlimbidx]

            t_hr = tmjd_hr[tidx]

            if t_hr_st_blend <= t_hr <= t_hr_ed_blend:
                if t_hr <= 8.0:
                    eofile_blend = imgindir + 'eovsa_{}.spw{}.tb.disk.fits'.format(
                        dateobj.strftime('%Y%m%d'), spwstr)
                    alpha = 1.0 - (t_hr - t_hr_st_blend) / (t_hr_ed_blend - t_hr_st_blend)

                else:
                    eofile_blend = imgindir_prevday + 'eovsa_{}.spw{}.tb.disk.fits'.format(
                        dateobj_prevday.strftime('%Y%m%d'), spwstr)
                    alpha = (t_hr - t_hr_st_blend) / (t_hr_ed_blend - t_hr_st_blend)
                alpha_blend = 1.0 - alpha
                eomap_blend = er.readfits(eofile_blend)
                eomap_blend = eomap_blend.resample(u.Quantity(eomap_blend.dimensions) / 2)
                eomap_blend.data[np.isnan(eomap_blend.data)] = 0.0
                eomap_rot_blend = diffrot_map(eomap_blend, time=timobj)
                offlimbidx = np.where(eomap_rot_blend.data == eomap_rot_blend.data[0, 0])
                eomap_rot_blend.data[offlimbidx] = eomap_blend.data[offlimbidx]
                eomap_rot_plt = smap.Map((eomap_rot.data * alpha + eomap_rot_blend.data * alpha_blend), eomap_rot.meta)
            else:
                eomap_rot_plt = eomap_rot

            # eomap_plt.plot(axes=ax, cmap=cmap, norm=norm)
            eomap_rot_plt.plot(axes=ax, cmap=cmap, norm=norm)

            ax.set_xlabel('')
            ax.set_ylabel('')
            ax.set_xticklabels([])
            ax.set_yticklabels([])
            ax.text(0.02, 0.02,
                    'EOVSA {:.1f} GHz  {}'.format(eomap.meta['CRVAL3'] / 1e9, dateobj.strftime('%d-%b-%Y %H:%M UT')),
                    transform=ax.transAxes, color='w', ha='left', va='bottom', fontsize=9)
            ax.text(0.98, 0.02, 'Max Tb {:.0f} K'.format(np.nanmax(eomap.data)),
                    transform=ax.transAxes, color='w', ha='right', va='bottom', fontsize=9)
            ax.set_xlim(-1227, 1227)
            ax.set_ylim(-1227, 1227)

            ax = axs[1]
            # for key, sourceid in aiaDataSource.items():
            ax.cla()

            if not os.path.exists(imgoutdir): os.makedirs(imgoutdir)
            sdourl = 'https://api.helioviewer.org/v2/getJP2Image/?date={}&sourceId={}'.format(timestr, sourceid)
            # print(sdourl)
            sdofile = os.path.join(imgoutdir, 'AIA' + key + '.{}.jp2'.format(tstrname))
            if not os.path.exists(sdofile):
                urllib.request.urlretrieve(sdourl, sdofile)

            if not os.path.exists(sdofile): continue
            sdomap = smap.Map(sdofile)
            norm = colors.Normalize()
            # sdomap_ = pmX.Sunmap(sdomap)
            if "HMI" in key:
                cmap = plt.get_cmap('gray')
            else:
                cmap = plt.get_cmap('sdoaia' + key.lstrip('0'))
            sdomap.plot(axes=ax, cmap=cmap, norm=norm)
            # sdomap_.imshow(axes=ax, cmap=cmap, norm=norm)
            # sdomap_.draw_limb(axes=ax, lw=0.25, alpha=0.5)
            # sdomap_.draw_grid(axes=ax, grid_spacing=10. * u.deg, lw=0.25)
            ax.set_xlabel('')
            ax.set_ylabel('')
            ax.set_xticklabels([])
            ax.set_yticklabels([])
            ax.text(0.02, 0.02,
                    '{}/{} {}  {}'.format(sdomap.observatory, sdomap.instrument.split(' ')[0], sdomap.measurement,
                                          sdomap.date.strftime('%d-%b-%Y %H:%M UT')),
                    transform=ax.transAxes, color='w', ha='left', va='bottom', fontsize=9)
            ax.set_xlim(-1227, 1227)
            ax.set_ylim(-1227, 1227)
            fig.savefig(figoutname, dpi=int(dpi), quality=85)
        imgfiles.append(figoutname)

    return imgfiles


def main(year, month, day=None, ndays=10, bd=3, show_warning=False):
    '''
    By default, the subroutine create EOVSA monthly movie
    '''
    if not show_warning:
        import warnings
        warnings.filterwarnings("ignore")
    # tst = datetime.strptime("2017-04-01", "%Y-%m-%d")
    # ted = datetime.strptime("2019-12-31", "%Y-%m-%d")
    if day is None:
        dst, ded = calendar.monthrange(year, month)
        tst = datetime(year, month, 1)
        ted = datetime(year, month, ded)
    else:
        if year:
            ted = datetime(year, month, day)
        else:
            ted = datetime.now() - timedelta(days=2)
        tst = Time(np.fix(Time(ted).mjd) - ndays, format='mjd').datetime
    tsep = datetime.strptime('2019-02-22', "%Y-%m-%d")

    vmaxs = [70.0e4, 35e4, 22e4, 16e4, 10e4, 8e4, 8e4]
    vmins = [-18.0e3, -8e3, -4.8e3, -3.4e3, -2.1e3, -1.6e3, -1.6e3]
    plt.ioff()
    fig, axs = plt.subplots(nrows=1, ncols=2, figsize=(8, 4))
    fig.subplots_adjust(bottom=0.0, top=1.0, left=0.0, right=1.0, hspace=0.0, wspace=0.0)

    dateobs = tst
    imgfileslist = []
    while dateobs < ted:
        monthstrdir = dateobs.strftime("%Y/%m/")
        imgoutdir = imgfitstmpdir + monthstrdir
        movieoutdir = pltfigdir + monthstrdir
        for odir in [imgoutdir, movieoutdir]:
            if not os.path.exists(odir):
                os.makedirs(odir)

        # dateobs = datetime.strptime("2019-12-21", "%Y-%m-%d")

        if dateobs > tsep:
            spws = ['0~1', '2~5', '6~10', '11~20', '21~30', '31~43', '44~49']
        else:
            spws = ['1~3', '4~9', '10~16', '17~24', '25~30']
        # spw = spws[bd:bd + 1]
        # vmax = vmaxs[bd:bd + 1]
        # vmin = vmins[bd:bd + 1]
        aiawave = '0304'

        tdateobs = Time(Time(dateobs).mjd + np.arange(0, 24, 1) / 24, format='mjd')
        imgfiles = pltEovsaQlookImageSeries(tdateobs, spws, vmaxs, vmins, aiawave, bd, fig=fig, axs=axs, overwrite=False,
                                            imgoutdir=imgoutdir)
        imgfileslist = imgfileslist + imgfiles
        dateobs = dateobs + timedelta(days=1)
    moviename = 'eovsa_bd{:02d}_aia304_{}'.format(bd + 1, dateobs.strftime("%Y%m"))
    DButil.img2movie(imgprefix=imgfileslist, outname=moviename, img_ext='jpg')
    os.system('mv {} {}'.format(os.path.join(imgoutdir + '../', moviename + '.mp4'),
                                os.path.join(movieoutdir, moviename + '.mp4')))
    plt.close('all')


# if __name__ == '__main__':
#     import sys
#     import numpy as np
#
#     # import subprocess
#     # shell = subprocess.check_output('echo $0', shell=True).decode().replace('\n', '').split('/')[-1]
#     # print("shell " + shell + " is using")
#
#     print(sys.argv)
#     try:
#         argv = sys.argv[1:]
#         if '--clearcache' in argv:
#             clearcache = True
#             argv.remove('--clearcache')  # Allows --clearcache to be either before or after date items
#         else:
#             clearcache = False
#
#         year = int(argv[0])
#         month = int(argv[1])
#         day = int(argv[2])
#         if len(argv) == 3:
#             dayspan = 30
#         else:
#             dayspan = int(argv[3])
#     except:
#         print('Error interpreting command line argument')
#         year = None
#         month = None
#         day = None
#         dayspan = 30
#         clearcache = True
#     print("Running pipeline_plt for date {}-{}-{}".format(year, month, day))
#     main(year, month, None, dayspan)


if __name__ == '__main__':
    '''
    Name: 
    eovsa_pltQlookMovie --- pipeline for plotting EOVSA daily full-disk image movie at multi frequencies.

    Synopsis:
    eovsa_pltQlookMovie.py [options]... [DATE_IN_YY_MM_DD]

    Description:
    Make EOVSA daily full-disk image movie at multi frequencies of the date specified
    by DATE_IN_YY_MM_DD (or from ndays before the DATE_IN_YY_MM_DD if option --ndays/-n is provided).
    If DATE_IN_YY_MM_DD is omitted, it will be set to 2 days before now by default. 
    The are no mandatory arguments in this command.


    -n, --ndays
            Processing the date spanning from DATE_IN_YY_MM_DD-ndays to DATE_IN_YY_MM_DD. Default is 30                                


    Example: 
    eovsa_pltQlookMovie.py -n 20 2020 06 10
    '''

    import sys
    import numpy as np
    import getopt

    year = None
    month = None
    day = None
    ndays = 30
    show_warning = False
    try:
        argv = sys.argv[1:]
        opts, args = getopt.getopt(argv, "n:w:",
                                   ['ndays=', 'show_warning='])
        print(opts, args)
        for opt, arg in opts:
            if opt in ('-n', '--ndays'):
                ndays = int(arg)
            elif opt in ('-w', '--show_warning'):
                if arg in ['True', 'T', '1']:
                    show_warning = True
                elif arg in ['False', 'F', '0']:
                    show_warning = False
                else:
                    show_warning = np.bool(arg)
        nargs = len(args)
        if nargs == 3:
            year = int(args[0])
            month = int(args[1])
            day = int(args[2])
        else:
            year = None
            month = None
            day = None
    except getopt.GetoptError as err:
        print(err)
        print('Error interpreting command line argument')
        year = None
        month = None
        day = None
        ndays = 30
        show_warning = False

    print("Running pipeline_plt for date {}-{}-{}.".format(year, month, day))
    kargs = {'ndays': ndays}
    for k, v in kargs.items():
        print(k, v)

    main(year=year, month=month, day=day, ndays=ndays, show_warning=show_warning)
